import shutil
import subprocess
import tempfile
import wave
from pathlib import Path
from unittest.mock import patch

import numpy as np
import pytest

from waddle.audios.call_tools import (
    convert_all_files_to_wav,
    convert_to_mp3,
    convert_to_wav,
    ensure_sampling_rate,
    remove_noise,
    transcribe,
    transcribe_in_batches,
)

# Define test directory paths
TESTS_DIR_PATH = Path(__file__).resolve().parents[3] / "tests"
EP0_DIR_PATH = TESTS_DIR_PATH / "ep0"

# Save the original subprocess.run
_orig_run = subprocess.run


def subprocess_run_with_error(error_on=None):
    """Patch `subprocess.run` to raise an error only when `error_on` is found in the command."""

    def side_effect(cmd, *args, **kwargs):
        cmd_str = " ".join(str(arg) for arg in cmd) if isinstance(cmd, (list, tuple)) else str(cmd)
        if error_on and error_on in cmd_str:
            raise subprocess.CalledProcessError(1, cmd)
        return _orig_run(cmd, *args, **kwargs)  # Call the original subprocess.run

    return patch("subprocess.run", side_effect=side_effect)


def get_wav_duration(file_path: Path) -> float:
    """Returns the duration of a WAV file."""
    with wave.open(str(file_path), "r") as wav_file_path:
        frames = wav_file_path.getnframes()
        rate = wav_file_path.getframerate()
        return frames / float(rate)


def get_total_noise_amount(file_path: Path, threshold: int = 150) -> float:
    """Returns the total amount of noise in a WAV file."""
    with wave.open(str(file_path), "r") as wav_file_path:
        n_channels = wav_file_path.getnchannels()
        n_frames = wav_file_path.getnframes()
        audio_data = wav_file_path.readframes(n_frames)

        audio_array = np.frombuffer(audio_data, dtype=np.int16)
        audio_array = np.abs(audio_array)

        if n_channels > 1:
            audio_array = audio_array.reshape(-1, n_channels).mean(axis=1)

        noise_amount = np.sum(audio_array[audio_array < threshold])
        return noise_amount


def test_convert_to_mp3():
    """Test that an `.m4a` file is converted to `.mp3` format."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())

        convert_to_mp3(temp_m4a)

        expected_output = temp_m4a.with_suffix(".mp3")
        assert expected_output.exists()


def test_convert_to_mp3_with_output_path():
    """Test that an `.m4a` file is converted to `.mp3` format with a custom output path."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())

        output_mp3 = temp_dir_path / "output.mp3"
        convert_to_mp3(temp_m4a, output_mp3)

        assert output_mp3.exists()


def test_convert_to_mp3_existing_mp3():
    """Skip conversion when `.mp3` file already exists."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())

        output_mp3 = temp_dir_path / "output.mp3"
        output_mp3.write_text("Existing MP3 file")
        convert_to_mp3(temp_m4a, output_mp3)

        assert output_mp3.exists()
        assert output_mp3.read_text() == "Existing MP3 file"


def test_convert_to_mp3_existing_mp3_force():
    """Convert with force=True even if `.mp3` file already exists."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())

        output_mp3 = temp_dir_path / "output.mp3"
        output_mp3.write_text("Existing MP3 file")
        convert_to_mp3(temp_m4a, output_mp3, force=True)

        assert output_mp3.exists()
        assert len(output_mp3.read_bytes()) > 100


def test_convert_to_mp3_error():
    """Test when an error occurs during conversion."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())

        with subprocess_run_with_error("ffmpeg"):
            with pytest.raises(RuntimeError, match="Converting"):
                convert_to_mp3(temp_m4a)


def test_convert_to_wav():
    """Test that an `.m4a` file is converted to `.wav` format."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())

        convert_to_wav(temp_m4a)

        expected_output = temp_m4a.with_suffix(".wav")
        assert expected_output.exists()


def test_convert_to_wav_with_output_path():
    """Test that an `.m4a` file is converted to `.wav` format with a custom output path."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())

        output_wav = temp_dir_path / "output.wav"
        convert_to_wav(temp_m4a, output_wav)

        assert output_wav.exists()


def test_convert_all_files_to_wav():
    """Test that `.m4a` files are converted to `.wav` format."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())  # Copy file using pathlib

        convert_all_files_to_wav(temp_dir_path)

        expected_output = temp_m4a.with_suffix(".wav")
        assert expected_output.exists()


def test_convert_all_files_to_wav_existing_wav():
    """Test when `.m4a` exists but corresponding `.wav` is already present."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"

    if not m4a_file.exists() or not wav_file_path.exists():
        pytest.skip(f"Sample files {m4a_file} or {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_wav_path = temp_dir_path / wav_file_path.name
        temp_m4a.write_bytes(m4a_file.read_bytes())  # Copy .m4a
        temp_wav_path.write_bytes(wav_file_path.read_bytes())  # Copy .wav (same name)

        with patch("builtins.print") as mock_print:
            convert_all_files_to_wav(temp_dir_path)

            # Ensure `[INFO]` message is printed when skipping `.m4a`
            assert any("[INFO] Skipping" in call.args[0] for call in mock_print.call_args_list)

        # Since .wav already exists, .m4a should be skipped (not overwritten)
        assert temp_wav_path.exists()


def test_convert_all_files_to_wav_error():
    """Test when an error occurs during conversion."""
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"
    if not m4a_file.exists():
        pytest.skip(f"Sample file {m4a_file} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_m4a = temp_dir_path / m4a_file.name
        temp_m4a.write_bytes(m4a_file.read_bytes())  # Copy file

        with subprocess_run_with_error("ffmpeg"):
            with pytest.raises(RuntimeError, match="Converting"):
                convert_all_files_to_wav(temp_dir_path)


def test_ensure_sampling_rate():
    """Test that the sampling rate is correctly set."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_wav_path = temp_dir_path / wav_file_path.name
        temp_wav_path.write_bytes(wav_file_path.read_bytes())  # Copy file

        output_wav = temp_dir_path / "output.wav"
        ensure_sampling_rate(temp_wav_path, output_wav, target_rate=16000)

        assert output_wav.exists()
        assert get_wav_duration(output_wav) == pytest.approx(
            get_wav_duration(temp_wav_path), rel=0.1
        )


def test_ensure_sampling_rate_file_not_found():
    """Test `ensure_sampling_rate` when input file does not exist."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        fake_input = temp_dir_path / "non_existent.wav"
        output_wav = temp_dir_path / "output.wav"

        with pytest.raises(FileNotFoundError, match="Input file not found"):
            ensure_sampling_rate(fake_input, output_wav, target_rate=16000)


def test_ensure_sampling_rate_error():
    """Test `ensure_sampling_rate` when an error occurs during conversion."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        fake_input = temp_dir_path / "input.wav"
        output_wav = temp_dir_path / "output.wav"

        # Create an empty input file
        fake_input.touch()

        with subprocess_run_with_error("ffmpeg"):
            with pytest.raises(RuntimeError, match="Converting"):
                ensure_sampling_rate(fake_input, output_wav, target_rate=16000)


def test_remove_noise():
    """Test noise removal using DeepFilterNet."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_wav_path = temp_dir_path / wav_file_path.name
        temp_wav_path.write_bytes(wav_file_path.read_bytes())

        output_wav = temp_dir_path / "denoised.wav"

        remove_noise(temp_wav_path, output_wav)

        assert output_wav.exists()
        assert get_wav_duration(output_wav) == pytest.approx(
            get_wav_duration(temp_wav_path), rel=0.1
        )
        assert output_wav.read_bytes() != temp_wav_path.read_bytes()

        # To compare noise levels, ensure_sampling_rate is used to convert to 48kHz
        temp_wav_path_48k = temp_dir_path / "temp_48k.wav"
        ensure_sampling_rate(temp_wav_path, temp_wav_path_48k, target_rate=48000)
        assert get_total_noise_amount(output_wav) < get_total_noise_amount(temp_wav_path_48k), (
            "Noise not removed"
        )


def test_remove_noise_same_output_path():
    """Test noise removal when input and output paths are the same."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        wav_path = temp_dir_path / wav_file_path.name
        wav_path.write_bytes(wav_file_path.read_bytes())

        remove_noise(wav_path, wav_path)

        assert wav_path.exists()
        assert get_wav_duration(wav_path) == pytest.approx(get_wav_duration(wav_file_path), rel=0.1)
        assert wav_path.read_bytes() != wav_file_path.read_bytes()

        # To compare noise levels, ensure_sampling_rate is used to convert to 48kHz
        original_48k_path = temp_dir_path / "wav_48k.wav"
        ensure_sampling_rate(wav_file_path, original_48k_path, target_rate=48000)
        assert get_total_noise_amount(wav_path) < get_total_noise_amount(original_48k_path), (
            "Noise not removed"
        )


def test_remove_noise_file_not_found():
    """Test remove_noise when input file does not exist."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        fake_input = temp_dir_path / "non_existent.wav"
        output_wav = temp_dir_path / "output.wav"

        with pytest.raises(FileNotFoundError, match="Input file not found"):
            remove_noise(fake_input, output_wav)


def test_remove_noise_error():
    """Test remove_noise when subprocess raises an error."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_wav_path = temp_dir_path / wav_file_path.name
        temp_wav_path.write_bytes(wav_file_path.read_bytes())

        output_wav = temp_dir_path / "denoised.wav"

        with subprocess_run_with_error("deep-filter"):
            with pytest.raises(RuntimeError, match="Running DeepFilterNet"):
                remove_noise(temp_wav_path, output_wav)


def test_remove_noise_with_missing_deep_filter():
    """Test that the DeepFilterNet installation command is executed when the tool is missing."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_wav_path = temp_dir_path / wav_file_path.name
        temp_wav_path.write_bytes(wav_file_path.read_bytes())

        with (
            patch(
                "waddle.audios.call_tools.install_deep_filter",
                side_effect=RuntimeError("install_deep_filter was called"),
            ),
            patch("waddle.audios.call_tools.get_tools_dir", return_value=temp_dir_path),
        ):
            with pytest.raises(RuntimeError, match="install_deep_filter was called"):
                remove_noise(temp_wav_path, temp_wav_path)


def test_transcribe():
    """Test transcription using Whisper."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_wav_path = temp_dir_path / wav_file_path.name
        temp_wav_path.write_bytes(wav_file_path.read_bytes())

        output_txt = temp_dir_path / "transcription.txt"

        transcribe(temp_wav_path, output_txt)

        assert output_txt.exists()
        assert len(output_txt.read_text().strip().split("\n")) >= 3


def test_transcribe_file_not_found():
    """Test transcribe when input file does not exist."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        fake_input = temp_dir_path / "non_existent.wav"
        output_txt = temp_dir_path / "transcription.txt"

        with pytest.raises(FileNotFoundError, match="Input file not found"):
            transcribe(fake_input, output_txt)


def test_transcribe_error():
    """Test transcribe when subprocess raises an error."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        temp_wav_path = temp_dir_path / wav_file_path.name
        temp_wav_path.write_bytes(wav_file_path.read_bytes())

        output_txt = temp_dir_path / "transcription.txt"

        with subprocess_run_with_error("whisper"):
            with pytest.raises(RuntimeError, match="Running Whisper"):
                transcribe(temp_wav_path, output_txt)


def test_transcribe_in_batches_1():
    """Test transcription in batches using Whisper."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        input_output_paths = []
        for i in range(5):
            temp_wav_path = temp_dir_path / f"audio_{i}.wav"
            shutil.copy(wav_file_path, temp_wav_path)
            input_output_paths.append((temp_wav_path, temp_dir_path / f"transcription_{i}.txt"))

        transcribe_in_batches(input_output_paths, batch_size=2)

        for _, output_path in input_output_paths:
            assert output_path.exists()
            assert len(output_path.read_text().strip().split("\n")) >= 3


def test_transcribe_in_batches_2():
    """Test transcription in batches using Whisper."""
    wav_file_path = EP0_DIR_PATH / "ep12-masa.wav"
    if not wav_file_path.exists():
        pytest.skip(f"Sample file {wav_file_path} not found")

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        input_output_paths = []
        for i in range(2):
            temp_wav_path = temp_dir_path / f"audio_{i}.wav"
            shutil.copy(wav_file_path, temp_wav_path)
            input_output_paths.append((temp_wav_path, temp_dir_path / f"transcription_{i}.txt"))

        # batch_size > len(input_output_paths)
        transcribe_in_batches(input_output_paths, batch_size=5)

        for _, output_path in input_output_paths:
            assert output_path.exists()
            assert len(output_path.read_text().strip().split("\n")) >= 3



import os
import re
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import Any

from mutagen.id3 import ID3

# FIXME: Importing from private(?) modules. This is not good but these items are
# not listed in `__all__` so can't be imported directly from the package as in the document.
# ref: https://mutagen.readthedocs.io/en/latest/user/id3.html#chapter-extension
from mutagen.id3._frames import CHAP, CTOC, TIT2
from mutagen.id3._specs import CTOCFlags

from waddle.audios.call_tools import convert_to_mp3


def generate_metadata(
    source_file: os.PathLike[Any] | str,
    audio_file: os.PathLike[Any] | str | None,
    output_dir: os.PathLike[Any] | str,
):
    source_file_path = Path(source_file)
    if not source_file_path.is_file():
        raise FileNotFoundError(f"Source file not found: {source_file}")

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    source = source_file_path.read_text(encoding="utf-8")
    entries = parse_annotated_srt(source)
    chapters, show_notes = extract_metadata(entries)

    base_name = source_file_path.stem
    chapter_file_path = Path(output_dir) / f"{base_name}.chapters.txt"
    show_notes_file_path = Path(output_dir) / f"{base_name}.show_notes.md"

    chapter_file_path.write_text(format_chapters(chapters), encoding="utf-8")
    show_notes_file_path.write_text(show_notes, encoding="utf-8")

    if audio_file is None:
        # check if the audio file with the same stem exists
        for ext in ["mp3", "m4a", "wav"]:
            audio_file_path = source_file_path.with_suffix(f".{ext}")
            if audio_file_path.is_file():
                print(f"[INFO] Found audio file: {audio_file_path}")
                audio_file = audio_file_path
                break
        else:
            print(f"[INFO] Audio file not found for: {source_file}")
            return

    audio_file_path = Path(audio_file)
    if not audio_file_path.is_file():
        raise FileNotFoundError(f"Audio file not found: {audio_file}")

    output_audio_path = output_dir / f"{audio_file_path.stem}.mp3"
    if audio_file_path.suffix == ".mp3":
        shutil.copy(audio_file_path, output_audio_path)
    else:
        convert_to_mp3(audio_file_path, output_audio_path)

    embed_chapter_info(output_audio_path, chapters)


@dataclass
class Chapter:
    start: float
    """Start time in seconds"""
    end: float
    """End time in seconds"""
    title: str
    """Chapter title"""
    level: int
    """Chapter level"""


@dataclass
class ChapterMarker:
    title: str
    level: int
    raw: str


@dataclass
class ShowNotesEntry:
    text: str


@dataclass
class SRTEntry:
    index: int
    """Subtitle index"""
    start: float
    """Start time in seconds"""
    end: float
    """End time in seconds"""
    speaker: str
    """Speaker name"""
    text: str
    """Subtitle text"""
    raw: str
    """Raw subtitle entry"""


def extract_metadata(
    entries: list[ChapterMarker | ShowNotesEntry | SRTEntry],
) -> tuple[list[Chapter], str]:
    chapters: list[Chapter] = []
    show_notes = ""

    # last SRT entry in this file
    chapter_start_srt_entry = None
    chapter_end_srt_entry = None
    for entry in reversed(entries):
        if isinstance(entry, SRTEntry):
            if chapter_end_srt_entry is None:
                chapter_end_srt_entry = entry
            chapter_start_srt_entry = entry
        if isinstance(entry, ChapterMarker):
            if chapter_start_srt_entry and chapter_end_srt_entry:
                chapters.insert(
                    0,
                    Chapter(
                        start=chapter_start_srt_entry.start,
                        end=chapter_end_srt_entry.end,
                        title=entry.title,
                        level=entry.level,
                    ),
                )
            chapter_start_srt_entry = None
            chapter_end_srt_entry = None
        if isinstance(entry, ShowNotesEntry):
            show_notes = entry.text + "\n" + show_notes

    return chapters, show_notes


def parse_annotated_srt(annotated_srt: str) -> list[ChapterMarker | ShowNotesEntry | SRTEntry]:
    s = annotated_srt
    items: list[ChapterMarker | ShowNotesEntry | SRTEntry] = []

    while s != "":
        # try to match an SRT entry
        srt_entry = match_srt_entry(s)
        if srt_entry:
            items.append(srt_entry)
            s = s[len(srt_entry.raw) :]
            continue
        # then try to match a chapter marker
        chapter_marker = match_chapter_marker(s)
        if chapter_marker:
            items.append(chapter_marker)
            s = s[len(chapter_marker.raw) :]
            continue
        # treat just the first line as show notes
        first_newline = s.find("\n")
        if first_newline != -1:
            show_notes_line = s[:first_newline]
            if show_notes_line.strip("\n ") != "":
                if show_notes_line.startswith(";"):
                    show_notes_line = show_notes_line[1:].strip()
                items.append(ShowNotesEntry(text=show_notes_line))
            s = s[first_newline + 1 :]
        else:
            # If there's no newline, treat all remaining text as show notes
            items.append(ShowNotesEntry(text=s))
            s = ""
    return items


srt_pattern = re.compile(
    r"^(\d+)\n(\d{2}:\d{2}:\d{2}\.\d{3}) --> (\d{2}:\d{2}:\d{2}\.\d{3})\n([^:]+): (.+)(?:\n\n|$)"
)


def match_srt_entry(s: str) -> SRTEntry | None:
    m = srt_pattern.match(s)
    if not m:
        return None
    return SRTEntry(
        index=int(m.group(1)),
        start=parse_time(m.group(2)),
        end=parse_time(m.group(3)),
        speaker=m.group(4),
        text=m.group(5),
        raw=m.group(0),
    )


def parse_time(ts: str) -> float:
    """Convert timestamp string to seconds."""
    h, m, s = map(float, ts.split(":"))
    return h * 3600 + m * 60 + s


chapter_pattern = re.compile(r"^(#{1,6}) (.+)\n")


def match_chapter_marker(s: str) -> ChapterMarker | None:
    m = chapter_pattern.match(s)
    if not m:
        return None
    level = len(m.group(1))
    title = m.group(2).strip()
    return ChapterMarker(title=title, level=level, raw=m.group(0))


def format_time(seconds: float) -> str:
    if seconds >= 3600:
        return f"{int(seconds // 3600):02d}:{int(seconds // 60 % 60):02d}:{int(seconds % 60):02d}"
    return f"{int(seconds // 60):02d}:{int(seconds % 60):02d}"


def format_chapters(chapters: list[Chapter]) -> str:
    return "\n".join(f"- ({format_time(chapter.start)}) {chapter.title}" for chapter in chapters)


def embed_chapter_info(mp3_file: Path, chapters: list[Chapter]):
    print(f"[INFO] Embedding chapter information in: {mp3_file}")
    audio = ID3(str(mp3_file))

    # Remove existing chapter information
    audio.delall("CTOC")
    audio.delall("CHAP")

    audio.add(
        CTOC(
            element_id="toc",
            flags=CTOCFlags.TOP_LEVEL | CTOCFlags.ORDERED,
            child_ids=[f"ch{idx + 1}".encode("utf-8") for idx, _ in enumerate(chapters)],
        )
    )

    for idx, chapter in enumerate(chapters):
        audio.add(
            CHAP(
                element_id=f"ch{idx + 1}".encode("utf-8"),
                start_time=int(chapter.start * 1000),
                end_time=int(chapter.end * 1000),
                sub_frames=[
                    TIT2(text=chapter.title),
                ],
            )
        )
    audio.save()



from pathlib import Path

import librosa
import numpy as np
import soundfile as sf
from scipy import signal

from waddle.config import (
    DEFAULT_COMP_AUDIO_DURATION,
    DEFAULT_SR,
)


def find_offset_via_cross_correlation(ref_audio: np.ndarray, spk_audio: np.ndarray) -> int:
    """
    Return the sample offset (lag) that best aligns spk_audio to ref_audio
    using cross-correlation.

    Positive offset => spk_audio starts earlier than ref_audio
    Negative offset => spk_audio starts later than ref_audio
    """
    correlation = signal.correlate(ref_audio, spk_audio, mode="full")
    max_corr_index = int(np.argmax(correlation))
    offset = max_corr_index - (
        len(spk_audio) - 1
    )  # the center index (zero lag) is at len(spk_audio) - 1
    return offset


def shift_audio(spk_audio: np.ndarray, offset: int, ref_length: int) -> np.ndarray:
    """
    Shift the spk_audio by 'offset' samples relative to the reference track.
    Ensures the returned array is the same length as the reference.
    """
    if offset > 0:
        shifted = np.pad(spk_audio, (offset, 0), mode="constant", constant_values=0)
    elif offset < 0:
        skip = abs(offset)
        if skip >= len(spk_audio):
            shifted = np.zeros(ref_length, dtype=spk_audio.dtype)
        else:
            shifted = spk_audio[skip:]
    else:
        shifted = spk_audio

    if len(shifted) > ref_length:
        shifted = shifted[:ref_length]
    elif len(shifted) < ref_length:
        shifted = np.pad(
            shifted, (0, ref_length - len(shifted)), mode="constant", constant_values=0
        )

    return shifted


def align_speaker_to_reference(
    reference_path: Path,
    speaker_path: Path,
    output_dir: Path = Path("out"),
    sample_rate: int = DEFAULT_SR,
    comp_duration: float = DEFAULT_COMP_AUDIO_DURATION,
) -> Path:
    # 1) Load short segments for cross-correlation
    ref_audio, _ = librosa.load(
        str(reference_path), sr=sample_rate, mono=True, duration=comp_duration
    )
    spk_audio, _ = librosa.load(
        str(speaker_path), sr=sample_rate, mono=True, duration=comp_duration
    )

    # Normalize
    if np.max(np.abs(ref_audio)) > 0:
        ref_audio /= np.max(np.abs(ref_audio))
    if np.max(np.abs(spk_audio)) > 0:
        spk_audio /= np.max(np.abs(spk_audio))

    # 2) Compute offset
    offset = find_offset_via_cross_correlation(ref_audio, spk_audio)

    # 3) Load full audio
    ref_full, _ = librosa.load(str(reference_path), sr=sample_rate, mono=True)
    spk_full, _ = librosa.load(str(speaker_path), sr=sample_rate, mono=True)

    # 4) Shift speaker
    aligned_speaker = shift_audio(spk_full, offset, len(ref_full))

    # 5) Save the aligned track
    output_dir.mkdir(parents=True, exist_ok=True)
    speaker_base = speaker_path.name
    aligned_path = output_dir / speaker_base
    sf.write(aligned_path, aligned_speaker, sample_rate)
    return aligned_path



from waddle.processing.combine import adjust_pos_to_timeline, merge_timelines


def test_adjust_pos_to_timeline_starting_from_zero():
    """Test adjust_pos_to_timeline where segments start from 0."""
    segments = [(0, 200), (300, 500)]
    assert adjust_pos_to_timeline(segments, 0) == 0, "Boundary adjustment at 0 failed."
    assert adjust_pos_to_timeline(segments, 100) == 100, "Boundary adjustment at 100 failed."
    assert adjust_pos_to_timeline(segments, 199) == 199, "Boundary adjustment at 199 failed."
    assert adjust_pos_to_timeline(segments, 200) == 200, "Boundary adjustment at 200 failed."
    assert adjust_pos_to_timeline(segments, 250) == 200, "Boundary adjustment at 250 failed."
    assert adjust_pos_to_timeline(segments, 300) == 200, "Boundary adjustment at 300 failed."
    assert adjust_pos_to_timeline(segments, 350) == 250, "Boundary adjustment at 350 failed."
    assert adjust_pos_to_timeline(segments, 500) == 400, "Boundary adjustment at 500 failed."


def test_adjust_pos_to_timeline_starting_from_nonzero():
    segments = [(100, 300), (400, 500)]
    assert adjust_pos_to_timeline(segments, 0) == 0, "Boundary adjustment at 0 failed."
    assert adjust_pos_to_timeline(segments, 99) == 0, "Boundary adjustment at 99 failed."
    assert adjust_pos_to_timeline(segments, 100) == 0, "Boundary adjustment at 100 failed."
    assert adjust_pos_to_timeline(segments, 299) == 199, "Boundary adjustment at 299 failed."
    assert adjust_pos_to_timeline(segments, 300) == 200, "Boundary adjustment at 300 failed."
    assert adjust_pos_to_timeline(segments, 350) == 200, "Boundary adjustment at 350 failed."
    assert adjust_pos_to_timeline(segments, 400) == 200, "Boundary adjustment at 400 failed."
    assert adjust_pos_to_timeline(segments, 450) == 250, "Boundary adjustment at 450 failed."
    assert adjust_pos_to_timeline(segments, 500) == 300, "Boundary adjustment at 500 failed."
    assert adjust_pos_to_timeline(segments, 501) == 300, "Boundary adjustment at 501 failed."


def test_adjust_pos_to_timeline_one_item():
    segments = [(0, 100)]
    assert adjust_pos_to_timeline(segments, 0) == 0, "Boundary adjustment at 0 failed."
    assert adjust_pos_to_timeline(segments, 50) == 50, "Boundary adjustment at 50 failed."
    assert adjust_pos_to_timeline(segments, 100) == 100, "Boundary adjustment at 100 failed."
    assert adjust_pos_to_timeline(segments, 101) == 100, "Boundary adjustment at 101 failed."


def test_merge_timelines_separated_segments():
    """Test merging timelines where segments are completely separate."""
    segments = [
        [(0, 100)],
        [(200, 300)],
    ]
    expected = [(0, 100), (200, 300)]
    result = merge_timelines(segments)
    assert result == expected, f"Expected {expected}, but got {result}"


def test_merge_timelines_continuous_segments():
    """Test merging timelines where segments are directly connected."""
    segments = [
        [(0, 100)],
        [(100, 200)],
        [(200, 300)],
    ]
    expected = [(0, 300)]
    result = merge_timelines(segments)
    assert result == expected, f"Expected {expected}, but got {result}"


def test_merge_timelines_non_overlapping_segments():
    """Test merging timelines with non-overlapping segments."""
    segments = [
        [(0, 100)],
        [(200, 300)],
        [(400, 500)],
    ]
    expected = [(0, 100), (200, 300), (400, 500)]
    result = merge_timelines(segments)
    assert result == expected, f"Expected {expected}, but got {result}"


def test_merge_timelines_overlapping_segments():
    """Test merging timelines where segments overlap."""
    segments = [
        [(200, 400)],
        [(0, 300)],
    ]
    expected = [(0, 400)]
    result = merge_timelines(segments)
    assert result == expected, f"Expected {expected}, but got {result}"



import argparse
from textwrap import dedent

from waddle.config import DEFAULT_COMP_AUDIO_DURATION, DEFAULT_LANGUAGE
from waddle.utils import phrase_time_to_seconds


def create_waddle_parser():
    """
    Create a command-line argument parser for the Waddle audio processing tool.

    The parser supports three subcommands:
    - single: Process a single audio file
    - preprocess: Preprocess multiple audio files
    - postprocess: Postprocess multiple audio files
    - metadata: Extract and process metadata from an annotated SRT file

    Returns:
        argparse.ArgumentParser: Configured argument parser
    """
    parser = argparse.ArgumentParser(exit_on_error=False)
    subparsers = parser.add_subparsers(dest="subcommand", required=True)

    # Create subparsers with descriptions
    single_parser = subparsers.add_parser(
        "single",
        description=(
            "Process a single audio file: "
            "noise removal, normalize, silence removal, and transcription."
        ),
    )
    preprocess_parser = subparsers.add_parser(
        "preprocess",
        description=(
            "Preprocess audio files: align audios, remove noise, normalize, and silence removal."
        ),
    )
    postprocess_parser = subparsers.add_parser(
        "postprocess",
        description="Postprocess audio files: silence removal, merge, and transcription.",
    )

    # ===== PRIMARY INPUT ARGUMENTS =====
    # Add single parser's input argument
    single_parser.add_argument(
        "audio",
        help="Path to the single audio file to process.",
    )

    # Add directory input for multi-file operations
    for p in [preprocess_parser, postprocess_parser]:
        p.add_argument(
            "-d",
            "--directory",
            default="./",
            help="Directory containing audio files (default: './').",
        )

    # ===== COMMON ARGUMENTS FOR ALL PARSERS =====
    parsers = [single_parser, preprocess_parser, postprocess_parser]
    for p in parsers:
        p.add_argument(
            "-o",
            "--output",
            default="./out",
            help="Directory to save the output (default: './out').",
        )
        p.add_argument(
            "-ss",
            type=phrase_time_to_seconds,
            default=0.0,
            help="Start time in seconds for the audio segment (default: 0.0).",
        )
        p.add_argument(
            "-t",
            "--time",
            type=phrase_time_to_seconds,
            default=None,
            help="Duration in seconds for the output audio (default: None).",
        )
        p.add_argument(
            "-wo",
            "--whisper-options",
            default=f"-l {DEFAULT_LANGUAGE}",
            help=(
                f"Options to pass to Whisper transcription (default: '-l {DEFAULT_LANGUAGE}').\n"
                "You can change the default language by modifying src/config.py."
            ),
        )

    # ===== OTHER OPTIONS =====
    # Noise removal (for audio preprocessing)
    for p in [single_parser, preprocess_parser]:
        p.add_argument(
            "-nnr",
            "--no-noise-remove",
            action="store_true",
            help="Skip removing noise from the audio.",
        )

    # preprocess options
    preprocess_parser.add_argument(
        "-r",
        "--reference",
        default=None,
        help="Path to the reference audio file (used in multi-file mode).",
    )
    preprocess_parser.add_argument(
        "-c",
        "--comp-duration",
        type=float,
        default=DEFAULT_COMP_AUDIO_DURATION,
        help=(
            "Duration in seconds for alignment comparison "
            f"(default: {DEFAULT_COMP_AUDIO_DURATION}s)."
        ),
    )
    preprocess_parser.add_argument(
        "-nc",
        "--no-convert",
        action="store_true",
        help="Skip converting audio files to WAV format.",
    )
    preprocess_parser.add_argument(
        "-tr",
        "--transcribe",
        action="store_true",
        help="Transcribe the processed audio files.",
    )

    # metadata
    show_notes_parser = subparsers.add_parser(
        "metadata",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=dedent("""\
            Generate metadata from an annotated SRT file.

            - `# Chapter` markers are used to define chapters.
                - Similar to Markdown, `#` indicates the level of the chapter, up to 6 levels.
                - The chapter starts at the next SRT timestamp and ends before next chapter.
            - Any other text is considered show notes.
            - Empty lines are ignored.
                - Use `;` to add newlines in show notes. The `;` will be deleted.
        """),
    )
    show_notes_parser.add_argument(
        "source",
        help="Path to the annotated SRT file.",
    )
    show_notes_parser.add_argument(
        "-i",
        "--input",
        help="Path to the input audio file. If not specified, it will look for an audio file with"
        " the same name.",
        default=None,
    )
    show_notes_parser.add_argument(
        "-o",
        "--output",
        help="Directory to save the metadata and audio files.",
        default="./metadata",
    )

    return parser



import tempfile
import wave
from pathlib import Path

import numpy as np
from pydub import AudioSegment
from scipy.io import wavfile

from waddle.processing.combine import (
    combine_audio_files,
)


def test_combine_audio_files_single_file():
    """Test combining when only a single file exists."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)
        audio_file = temp_dir / "single.wav"
        output_audio = temp_dir / "output.wav"

        AudioSegment.silent(duration=1000).export(str(audio_file), format="wav")

        combine_audio_files([audio_file], output_audio)

        assert output_audio.exists(), "Output audio file was not created."
        with wave.open(str(output_audio), "r") as wf:
            assert wf.getnframes() > 0, "Output audio file is empty."


def generate_sine_wave(frequency=440, duration_ms=500, sample_rate=44100, amplitude=0.5):
    """
    Generate a NumPy array representing a sine wave.
    """
    t = np.linspace(0, duration_ms / 1000, int(sample_rate * (duration_ms / 1000)), endpoint=False)
    waveform = (amplitude * np.sin(2 * np.pi * frequency * t) * 32767).astype(np.int16)
    return waveform


def test_combine_audio_files_with_numpy_verification():
    """
    Test combining three different sine wave audio segments and verify output using NumPy.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir = Path(temp_dir)
        sample_rate = 44100
        duration_ms = 500  # 500ms per segment

        # Generate three different sine wave signals
        segment1 = generate_sine_wave(
            frequency=440, duration_ms=duration_ms, sample_rate=sample_rate
        )
        segment2 = generate_sine_wave(
            frequency=880, duration_ms=duration_ms, sample_rate=sample_rate
        )
        segment3 = generate_sine_wave(
            frequency=1760, duration_ms=duration_ms, sample_rate=sample_rate
        )

        # Save the three segments as separate WAV files
        paths = []
        for i, segment in enumerate([segment1, segment2, segment3]):
            path = temp_dir / f"segment_{i}.wav"
            wavfile.write(str(path), sample_rate, segment)
            paths.append(str(path))

        # Output path for combined audio
        output_audio_path = temp_dir / "combined.wav"

        # Combine audio files
        combine_audio_files(paths, output_audio_path)

        # Read the output audio file
        output_sample_rate, output_audio = wavfile.read(str(output_audio_path))

        # Ensure the sample rate is unchanged
        assert output_sample_rate == sample_rate, "Sample rate mismatch in output audio."

        # Expected output: Element-wise sum of the three signals
        expected_audio = (
            segment1.astype(np.int32) + segment2.astype(np.int32) + segment3.astype(np.int32)
        )

        # Prevent overflow by clipping to int16 range
        expected_audio = np.clip(expected_audio, -32768, 32767).astype(np.int16)

        # Compare the actual combined audio with the expected sum
        np.testing.assert_array_equal(
            output_audio, expected_audio, "Combined audio does not match expected waveform."
        )



import shutil
from pathlib import Path

import numpy as np
from pydub import AudioSegment
from tqdm import tqdm

from waddle.audios.call_tools import transcribe_in_batches
from waddle.config import (
    DEFAULT_BUFFER_DURATION,
    DEFAULT_CHUNK_DURATION,
    DEFAULT_LANGUAGE,
    DEFAULT_TARGET_DB,
    DEFAULT_THRESHOLD_DB,
)
from waddle.processing.combine import SpeechTimeline, combine_segments_into_audio
from waddle.utils import format_audio_filename, format_time, parse_audio_filename, time_to_seconds


def detect_speech_timeline(
    audio_path: Path,
    threshold_db: float = DEFAULT_THRESHOLD_DB,
    chunk_size_ms: int = int(DEFAULT_CHUNK_DURATION * 1000),
    buffer_size_ms: int = int(DEFAULT_BUFFER_DURATION * 1000),
    target_dBFS: float = DEFAULT_TARGET_DB,
) -> tuple[Path, SpeechTimeline]:
    """
    Detects speech segments in an audio file based on a specified loudness threshold.
    Each detected segment includes a buffer of audio before and after to ensure completeness.
    The detected segments are normalized to achieve a global mean dBFS of target_dBFS.

    Args:
        audio_path (str): Path to the input audio file.
        threshold_db (float): Loudness threshold in dBFS for detecting speech.
        chunk_size_ms (int): Duration of each audio chunk in milliseconds.
        buffer_size_ms (int): Additional buffer duration in milliseconds for segment merging.
        target_dBFS (float): Target loudness level (dBFS) for normalized audio segments.
        out_duration (float, optional): Maximum duration of the processed output audio in seconds.

    Returns
        segs_folder_path (str): Path to the directory containing the extracted speech segments.
        merged_segments (SpeechTimeline): List of detected and merged speech segments.
    """
    audio = AudioSegment.from_file(str(audio_path))

    segments = []
    current_segment = None

    # Create a clean 'chunks' folder
    identifier = audio_path.stem
    chunks_folder = audio_path.parent / "chunks" / identifier
    if chunks_folder.exists():
        shutil.rmtree(chunks_folder)
    chunks_folder.mkdir(parents=True, exist_ok=True)

    duration = len(audio)
    for i in tqdm(
        range(0, duration, chunk_size_ms),
        desc="[INFO] Detecting speech segments",
    ):
        chunk = audio[i : i + chunk_size_ms]
        if chunk.dBFS > threshold_db:
            start_ms = max(0, i - buffer_size_ms)
            end_ms = min(duration, i + chunk_size_ms + buffer_size_ms)
            if current_segment is None:
                current_segment = [start_ms, end_ms]
            else:
                current_segment[1] = end_ms
        else:
            if current_segment is not None:
                segments.append((current_segment[0], current_segment[1]))
                current_segment = None

    # Finalize last segment if any
    if current_segment is not None:
        segments.append((current_segment[0], current_segment[1]))

    # Clean up chunks
    shutil.rmtree(chunks_folder)

    # Merge overlapping/adjacent segments
    merged_segments = merge_segments(segments)

    # Save segment to disk
    audio_file_name = audio_path.stem
    segs_folder_path = audio_path.parent / f"{audio_file_name}_segs"
    if segs_folder_path.exists():
        shutil.rmtree(segs_folder_path)
    segs_folder_path.mkdir(parents=True, exist_ok=True)

    # collect max_dBFS for each segment
    max_dBFS_list = []
    for seg in merged_segments:
        seg_audio = audio[seg[0] : seg[1]]
        max_dBFS_list.append(seg_audio.dBFS)

    # calculate 95th percentile of max_dBFS
    if not max_dBFS_list:
        print("[Warning] No speech segments detected.")
        return segs_folder_path, []
    max_dBFS_95th_percentile = np.percentile(max_dBFS_list, 95)

    # Calculate gain adjustment to achieve target_dBFS for 95th percentile
    gain_adjustment = target_dBFS - max_dBFS_95th_percentile

    # Apply normalization to all segments
    for seg in merged_segments:
        seg_audio = audio[seg[0] : seg[1]]
        normalized_audio = seg_audio.apply_gain(gain_adjustment)
        seg_audio_path = segs_folder_path / format_audio_filename("seg", seg[0], seg[1])
        normalized_audio.export(seg_audio_path, format="wav")

    # Clean up audio
    audio_path.unlink()

    print(f"[INFO] Global normalization applied with gain adjustment: {gain_adjustment} dB")

    return segs_folder_path, merged_segments


def merge_segments(segments: SpeechTimeline) -> SpeechTimeline:
    merged_segments = []
    for seg in segments:
        if not merged_segments or seg[0] > merged_segments[-1][1]:
            merged_segments.append(seg)
        else:
            merged_segments[-1] = (
                merged_segments[-1][0],
                max(merged_segments[-1][1], seg[1]),
            )
    return merged_segments


def process_segments(
    segs_folder_path: Path,
    combined_audio_path: Path,
    transcription_path: Path | None = None,
    whisper_options: str = f"-l {DEFAULT_LANGUAGE}",
    timeline: SpeechTimeline | None = None,
) -> None:
    """
    Transcribe only the detected speech segments, adjust timestamps,
    and combine them into a single audio file.

    Args:
        segs_folder_path (str): Path to the folder containing the speech segments.
        combined_audio_path (str): Path to save the combined audio file.
        transcription_path (str): Path to save the combined transcription file.
        language (str): Language code for transcription.
    """
    seg_file_paths = sorted(segs_folder_path.glob("*.wav"), key=lambda x: int(x.stem.split("_")[1]))
    transcription_entries = []

    if transcription_path:
        input_output_paths = []
        for segs_file_path in seg_file_paths:
            # Transcribe segment
            srt_output_path = Path(segs_file_path).with_suffix(".srt")
            input_output_paths.append((segs_file_path, srt_output_path))

        transcribe_in_batches(input_output_paths, options=whisper_options)

        for segs_file_path, srt_output_path in input_output_paths:
            start, _ = parse_audio_filename(str(segs_file_path))
            start_seconds = float(start) / 1000

            # Adjust transcription timestamps
            process_segment_transcription(
                segs_file_path.with_suffix(".srt"), start_seconds, transcription_entries
            )
            srt_output_path.unlink()

        # Create a single SRT file from all segments
        with open(str(transcription_path), "w", encoding="utf-8") as srt_out:
            for idx, (start_time, end_time, text) in enumerate(transcription_entries, start=1):
                srt_out.write(f"{idx}\n")
                srt_out.write(f"{start_time} --> {end_time}\n")
                srt_out.write(f"{text}\n\n")

    # Combine segments into one audio file
    combine_segments_into_audio(
        segs_folder_path,
        combined_audio_path,
        timeline=timeline,
    )


def process_segment_transcription(
    transcribe_file_path: Path, start_offset: float, transcription_entries: list
) -> None:
    """
    Adjust timestamps in a segment's transcription file by adding the start offset,
    then append entries to the shared list.

    Args:
        transcribe_file_path (str): Path to the .srt transcription file.
        start_offset (float): Offset in seconds to add to each timestamp.
        transcription_entries (list): Output list for adjusted transcription entries.
    """
    transcribe_file_path = Path(
        transcribe_file_path
    )  # TODO: Delete it after switch to Pathlib in test
    if not transcribe_file_path.is_file():
        print(f"[Warning] SRT file not found for segment: {transcribe_file_path}")
        return

    with open(str(transcribe_file_path), "r", encoding="utf-8") as srt_file:
        blocks = srt_file.read().strip().split("\n\n")

    for block in blocks:
        lines = block.split("\n")
        if len(lines) < 3:
            continue

        # Extract timestamps and text
        _, timestamps, *text_lines = lines
        s_timestamp, e_timestamp = timestamps.split(" --> ")
        text = " ".join(text_lines)

        # Adjust timestamps
        adjusted_start = format_time(start_offset + time_to_seconds(s_timestamp))
        adjusted_end = format_time(start_offset + time_to_seconds(e_timestamp))

        transcription_entries.append((adjusted_start, adjusted_end, text))



import tempfile
import wave
from pathlib import Path

import numpy as np

from waddle.audios.align_offset import (
    align_speaker_to_reference,
    find_offset_via_cross_correlation,
    shift_audio,
)

DEFAULT_SR = 48000


def generate_test_audio(sr=DEFAULT_SR, s_padding=0, e_padding=0):
    """
    Generate a test audio with s_padding seconds of silence, followed by 1 second of sine wave,
    followed by e_padding seconds of silence.
    """
    t = np.linspace(0, np.pi / 2, int(sr * 1.0), endpoint=False)
    sine_wave = (32767 * 0.5 * np.sin(t)).astype(np.float32)

    s_silence = np.zeros(sr + s_padding, dtype=np.float32)  # 1s + s_padding
    e_silence = np.zeros(sr + e_padding, dtype=np.float32)  # 1s + e_padding
    audio = np.concatenate([s_silence, sine_wave, e_silence])

    return audio, sr


def write_wav(file_path, audio, sr):
    """Write an audio file in WAV format."""
    with wave.open(str(file_path), "w") as wav_file:
        wav_file.setnchannels(1)
        wav_file.setsampwidth(2)
        wav_file.setframerate(sr)
        wav_file.writeframes(audio.tobytes())


def read_wav(file_path):
    """Read an audio file in WAV format."""
    with wave.open(str(file_path), "r") as wav_file:
        sr = wav_file.getframerate()
        audio = np.frombuffer(wav_file.readframes(wav_file.getnframes()), dtype=np.float32)
    return audio, sr


def test_find_offset_via_cross_correlation():
    ref_audio, _ = generate_test_audio()
    spk_audio, _ = generate_test_audio(s_padding=24000)

    offset = find_offset_via_cross_correlation(ref_audio, spk_audio)
    assert offset == -24000  # offset should be negative to align the speaker to the reference


def test_shift_audio_s_padding():
    ref_audio, _ = generate_test_audio()
    spk_audio, _ = generate_test_audio(s_padding=4800)
    shifted = shift_audio(spk_audio, -4800, len(ref_audio))
    assert np.allclose(ref_audio, shifted)


def test_shift_audio_e_padding():
    ref_audio, _ = generate_test_audio()
    spk_audio, _ = generate_test_audio(e_padding=158)
    shifted = shift_audio(spk_audio, 0, len(ref_audio))
    assert len(shifted) == len(ref_audio)
    assert np.allclose(ref_audio, shifted)


def test_shift_audio_both():
    ref_audio, _ = generate_test_audio()
    spk_audio, _ = generate_test_audio(s_padding=800, e_padding=555)
    shifted = shift_audio(spk_audio, -800, len(ref_audio))
    assert len(shifted) == len(ref_audio)
    assert np.allclose(ref_audio, shifted)


def test_align_speaker_to_reference():
    with tempfile.TemporaryDirectory() as temp_dir:
        ref_audio, sr = generate_test_audio()
        spk_audio_0, _ = generate_test_audio(s_padding=4800, e_padding=100)
        spk_audio_1, _ = generate_test_audio(e_padding=800)

        temp_dir_path = Path(temp_dir)
        ref_path = temp_dir_path / "ref.wav"
        spk_path_0 = temp_dir_path / "spk_0.wav"
        spk_path_1 = temp_dir_path / "spk_1.wav"

        write_wav(ref_path, ref_audio, sr)
        write_wav(spk_path_0, spk_audio_0, sr)
        write_wav(spk_path_1, spk_audio_1, sr)

        output_path_0 = align_speaker_to_reference(ref_path, spk_path_0, temp_dir_path)
        output_path_1 = align_speaker_to_reference(ref_path, spk_path_1, temp_dir_path)

        aligned_audio_0, _ = read_wav(output_path_0)
        aligned_audio_1, _ = read_wav(output_path_1)

        assert len(aligned_audio_0) == len(ref_audio)
        assert len(aligned_audio_1) == len(ref_audio)

        assert np.allclose(ref_audio, aligned_audio_0)
        assert np.allclose(ref_audio, aligned_audio_1)



import shutil

from waddle.argparse import create_waddle_parser
from waddle.metadata import generate_metadata
from waddle.processor import postprocess_multi_files, preprocess_multi_files, process_single_file
from waddle.utils import to_path


def main():
    """
    Command-line entry point for processing audio files.
    """
    parser = create_waddle_parser()
    args = parser.parse_args()

    match args.subcommand:
        case "single":
            do_single(args)
        case "preprocess":
            do_preprocess(args)
        case "postprocess":
            do_postprocess(args)
        case "metadata":
            do_metadata(args)
        case _:
            raise ValueError(f"Command not implemented: {args.subcommand}")


def do_single(args):
    audio_path = to_path(args.audio)
    if not audio_path.is_file():
        raise FileNotFoundError(f"Audio file not found: {args.audio}")
    output_dir = to_path(args.output)
    output_dir.mkdir(parents=True, exist_ok=True)

    # Copy the audio file to the output directory
    audio_file_name = audio_path.name
    tmp_audio_path = output_dir / audio_file_name
    print(f"[INFO] Copying audio file to: {tmp_audio_path}")
    shutil.copy(args.audio, tmp_audio_path)

    print(f"[INFO] Processing single audio file: {tmp_audio_path}")
    process_single_file(
        aligned_audio=tmp_audio_path,
        output_dir=output_dir,
        speaker_audio=audio_path,
        ss=args.ss,
        out_duration=args.time,
        no_noise_remove=args.no_noise_remove,
        whisper_options=args.whisper_options,
    )
    print(f"[INFO] Processed single audio file saved in: {output_dir}")


def do_preprocess(args):
    reference_path_or_none = to_path(args.reference) if args.reference else None
    if reference_path_or_none is not None and not reference_path_or_none.is_file():
        raise FileNotFoundError(f"Reference file not found: {args.reference}")
    source_dir_path = to_path(args.directory)
    if not source_dir_path.is_dir():
        raise FileNotFoundError(f"Audio source directory not found: {args.directory}")
    output_dir_path = to_path(args.output or "./out")

    preprocess_multi_files(
        reference=reference_path_or_none,
        source_dir=source_dir_path,
        output_dir=output_dir_path,
        comp_duration=args.comp_duration,
        ss=args.ss,
        out_duration=args.time,
        no_noise_remove=args.no_noise_remove,
        convert=not args.no_convert,
        transcribe=args.transcribe,
        whisper_options=args.whisper_options,
    )


def do_postprocess(args):
    source_dir_path = to_path(args.directory)
    if not source_dir_path.is_dir():
        raise FileNotFoundError(f"Audio source directory not found: {args.directory}")
    output_dir_path = to_path(args.output or "./out")

    print(f"[INFO] Postprocessing audio files from: {source_dir_path}")
    postprocess_multi_files(
        source_dir=source_dir_path,
        output_dir=output_dir_path,
        ss=args.ss,
        out_duration=args.time,
        whisper_options=args.whisper_options,
    )
    print(f"[INFO] Postprocessing complete. Output saved in: {output_dir_path}")


def do_metadata(args):
    generate_metadata(args.source, args.input, args.output)


if __name__ == "__main__":
    main()



import tempfile
from pathlib import Path
from typing import Dict, TypeAlias

from waddle.processing.combine import SrtEntries, SrtEntry, combine_srt_files, parse_srt

SrtFiles: TypeAlias = dict[str, str]


def create_srt_files(temp_dir: Path, srt_files: SrtFiles, is_combined: bool = False) -> None:
    temp_dir = Path(temp_dir)
    for filename, content in srt_files.items():
        (temp_dir / f"{filename}.srt").write_text(content, encoding="utf-8")

    if is_combined:
        output_srt = temp_dir / "combined.srt"
        combine_srt_files(temp_dir, output_srt)
        assert output_srt.exists(), "Combined SRT file was not created."


def check_srt_entries(temp_dir: Path, expected_srt_files: Dict[str, SrtEntries]):
    temp_dir = Path(temp_dir)
    for filename, expected_entries in expected_srt_files.items():
        entries = parse_srt(temp_dir / f"{filename}.srt")
        assert len(entries) == len(expected_entries), "Number of entries mismatch."
        for i, entry in enumerate(entries):
            check_srt_entry(entry, expected_entries[i])


def check_srt_entry(entry: SrtEntry, expected_entry: SrtEntry):
    assert entry[0] == expected_entry[0], f"Start time mismatch: {entry[0]} != {expected_entry[0]}"
    assert entry[1] == expected_entry[1], f"End time mismatch: {entry[1]} != {expected_entry[1]}"
    assert entry[2] == expected_entry[2], f"Text mismatch: {entry[2]} != {expected_entry[2]}"


def test_parse_srt_empty_file():
    """Test parsing an empty SRT file."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        srt_files = {
            "empty": "",
        }
        create_srt_files(temp_dir_path, srt_files)

        expected_srt_files = {"empty": []}
        check_srt_entries(temp_dir_path, expected_srt_files)


def test_parse_srt_single_entry():
    """Test parsing an SRT file with a single entry."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        srt_files = {
            "single": ("1\n00:00:00,000 --> 00:00:05,000\nHello world.\n\n"),
        }
        create_srt_files(temp_dir_path, srt_files)

        expected_srt_files = {"single": [("00:00:00.000", "00:00:05.000", "Hello world.")]}
        check_srt_entries(temp_dir_path, expected_srt_files)


def test_parse_srt_multiple_entries():
    """Test parsing an SRT file with multiple entries."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        srt_files = {
            "multiple": (
                "1\n00:00:00,000 --> 00:00:05,000\nHello world.\n\n"
                "2\n00:00:05,000 --> 00:00:10,000\nHow are you?\n\n"
                "3\n00:00:10,000 --> 00:00:15,000\nI'm fine, thanks.\n\n"
            ),
        }
        create_srt_files(temp_dir_path, srt_files)

        expected_srt_files = {
            "multiple": [
                ("00:00:00.000", "00:00:05.000", "Hello world."),
                ("00:00:05.000", "00:00:10.000", "How are you?"),
                ("00:00:10.000", "00:00:15.000", "I'm fine, thanks."),
            ]
        }
        check_srt_entries(temp_dir_path, expected_srt_files)


def test_parse_srt_multiple_entries_with_broken():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        srt_files = {
            "multiple": (
                "1\n00:00:00,000 --> 00:00:05,000\nHello world.\n\n"
                "2\n00:00:05,000 --> 00:00:10,000\nHow are you?\n"  # Not 2 newlines
                "3\n00:00:15,000 --> 00:00:20,000\nI'm fine, thanks.\n\n"
                "4\n00:00:25,000 --> 00:00:30,000 What's up?\n\n"  # Missing newline after timestamp
                "5\n00:00:35,000 --> 00:00:40,000\nI'm good.\n\n"
            ),
        }
        create_srt_files(temp_dir_path, srt_files)

        expected_srt_files = {
            "multiple": [
                ("00:00:00.000", "00:00:05.000", "Hello world."),
                (
                    "00:00:05.000",
                    "00:00:10.000",
                    "How are you? 3 00:00:15,000 --> 00:00:20,000 I'm fine, thanks.",
                ),
                ("00:00:35.000", "00:00:40.000", "I'm good."),
            ]
        }
        check_srt_entries(temp_dir_path, expected_srt_files)


def test_combine_srt_files_empty_directory():
    """Test combining SRT files when the input directory is empty."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        output_srt_path = temp_dir_path / "combined.srt"

        combine_srt_files(temp_dir_path, output_srt_path)

        expected_srt_files = {"combined": []}
        check_srt_entries(temp_dir_path, expected_srt_files)


def test_combine_srt_files_single_file():
    """Test combining a single SRT file."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        srt_files = {
            "speaker1": "1\n00:00:00,000 --> 00:00:05,000\nHello world.\n\n",
        }
        create_srt_files(temp_dir_path, srt_files, is_combined=True)

        expected_srt_files = {
            "combined": [("00:00:00.000", "00:00:05.000", "speaker1: Hello world.")]
        }
        check_srt_entries(temp_dir_path, expected_srt_files)


def test_combine_srt_files_multiple_files():
    """Test combining multiple SRT files and sorting by timestamps."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        srt_files = {
            "speaker1": "1\n00:00:05,000 --> 00:00:10,000\nHow are you?\n\n",
            "ep0-speaker2": (  # hyphenated name
                "1\n00:00:00,000 --> 00:00:05,000\nHello world.\n\n"
                "2\n00:00:15,000 --> 00:00:20,000\nWhat's up?\n\n"
            ),
            "speaker3": (
                "1\n00:00:10,000 --> 00:00:15,000\nI'm fine, thanks.\n\n"
                "2\n00:00:20,000 --> 00:00:25,000\nI'm good.\n\n"
            ),
        }
        create_srt_files(temp_dir_path, srt_files, is_combined=True)

        expected_srt_files = {
            "combined": [
                ("00:00:00.000", "00:00:05.000", "speaker2: Hello world."),
                ("00:00:05.000", "00:00:10.000", "speaker1: How are you?"),
                ("00:00:10.000", "00:00:15.000", "speaker3: I'm fine, thanks."),
                ("00:00:15.000", "00:00:20.000", "speaker2: What's up?"),
                ("00:00:20.000", "00:00:25.000", "speaker3: I'm good."),
            ]
        }
        check_srt_entries(temp_dir_path, expected_srt_files)


def test_combine_srt_files_with_broken_and_empty_files():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        srt_files = {
            "ep0-speaker1": "1\n00:00:05,000 --> 00:00:10,000\nI'm fine, thanks.\n\n",
            "broken": (
                "1\n00:00:07,000 --> 00:00:08,000\nI'm good too.\n\n"
                "2\n00:00:01,000 --> 00:00:30,000 Missing newline\n\n"  # Missing newline after text
                "3\n00:00:10,001 --> 00:00:40,000\nThis is next of missing newline.\n\n"
            ),
            "empty": "",
            "speaker2": "1\n00:00:00,000 --> 00:00:06,000\nHow are you?\n\n",
        }
        create_srt_files(temp_dir_path, srt_files, is_combined=True)

        expected_srt_files = {
            "combined": [
                ("00:00:00.000", "00:00:06.000", "speaker2: How are you?"),
                ("00:00:05.000", "00:00:10.000", "speaker1: I'm fine, thanks."),
                (
                    "00:00:07.000",
                    "00:00:08.000",
                    "broken: I'm good too.",
                ),
                (
                    "00:00:10.001",
                    "00:00:40.000",
                    "broken: This is next of missing newline.",
                ),
            ]
        }
        check_srt_entries(temp_dir_path, expected_srt_files)



import wave
from pathlib import Path


def clip_audio(
    audio_path: Path, out_path: Path, ss: float = 0.0, out_duration: float | None = None
) -> Path:
    """
    Clip an audio file to a specified duration starting from a given time.

    Args:
        audio_path (str): Path to the input audio file.
        ss (float): Start time in seconds (default: 0.0).
        out_duration (float | None): Duration in seconds for the output audio (default: None).

    Returns:
        str: Path to the clipped audio file.
    """
    with wave.open(str(audio_path), "rb") as wav:
        frame_rate = wav.getframerate()
        n_channels = wav.getnchannels()
        samp_width = wav.getsampwidth()

        start_frame = int(ss * frame_rate)
        if out_duration is None:
            end_frame = wav.getnframes()
        else:
            end_frame = start_frame + int(out_duration * frame_rate)

        wav.setpos(start_frame)
        frames = wav.readframes(end_frame - start_frame)

    with wave.open(str(out_path), "wb") as output_wav:
        output_wav.setnchannels(n_channels)
        output_wav.setsampwidth(samp_width)
        output_wav.setframerate(frame_rate)
        output_wav.writeframes(frames)

    return audio_path



import tempfile
import wave
from pathlib import Path

from waddle.audios.clip import clip_audio

TESTS_DIR_PATH = Path(__file__).resolve().parent.parents[2] / "tests"
EP0_DIR_PATH = TESTS_DIR_PATH / "ep0"


def get_wav_duration(filename):
    """Returns the duration of a WAV file."""
    with wave.open(filename, "r") as wav_file:
        frames = wav_file.getnframes()
        rate = wav_file.getframerate()
        duration = frames / float(rate)
        return duration


def test_clip_audio_01():
    with tempfile.NamedTemporaryFile() as temp_file:
        clip_audio(EP0_DIR_PATH / "ep12-masa.wav", Path(temp_file.name), 0, 10)
        assert get_wav_duration(temp_file.name) == 10  # 10 seconds


def test_clip_audio_02():
    with tempfile.NamedTemporaryFile() as temp_file:
        clip_audio(EP0_DIR_PATH / "ep12-masa.wav", Path(temp_file.name), 0, 30)
        assert (
            14 <= get_wav_duration(temp_file.name) <= 15
        )  # Should match the original length (14.052s)


def test_clip_audio_03():
    with tempfile.NamedTemporaryFile() as temp_file:
        clip_audio(EP0_DIR_PATH / "ep12-masa.wav", Path(temp_file.name), 5, 30)
        assert (
            9 <= get_wav_duration(temp_file.name) <= 10
        )  # should be original length (14.052) - 5 seconds


def test_clip_audio_04():
    with tempfile.NamedTemporaryFile() as temp_file:
        clip_audio(EP0_DIR_PATH / "ep12-masa.wav", Path(temp_file.name), 5)
        assert 9 <= get_wav_duration(temp_file.name) <= 10  # same as above



import platform
import sys
import urllib.request
from pathlib import Path

from platformdirs import user_runtime_dir

from waddle.config import APP_AUTHOR, APP_NAME


def install_deep_filter():
    # Tool installation directories
    TOOLS_DIR = Path(user_runtime_dir(APP_NAME, APP_AUTHOR)) / "tools"
    DEEP_FILTER_OUTPUT = TOOLS_DIR / "deep-filter"
    if DEEP_FILTER_OUTPUT.exists():
        print(f"DeepFilterNet binary already exists: {DEEP_FILTER_OUTPUT}")
        return

    DEEP_FILTER_VERSION = "0.5.6"
    DEEP_FILTER_BASE_URL = (
        f"https://github.com/Rikorose/DeepFilterNet/releases/download/v{DEEP_FILTER_VERSION}"
    )

    # Create the tools directory if it doesn't exist
    TOOLS_DIR.mkdir(parents=True, exist_ok=True)

    # Detect system architecture and platform
    ARCH = platform.machine().lower()
    OS = platform.system().lower()

    # Determine the correct binary for DeepFilterNet
    ARCH_OS_MAP = {
        ("aarch64", "darwin"): f"deep-filter-{DEEP_FILTER_VERSION}-aarch64-apple-darwin",
        ("arm64", "darwin"): f"deep-filter-{DEEP_FILTER_VERSION}-aarch64-apple-darwin",
        ("aarch64", "linux"): f"deep-filter-{DEEP_FILTER_VERSION}-aarch64-unknown-linux-gnu",
        ("arm64", "linux"): f"deep-filter-{DEEP_FILTER_VERSION}-aarch64-unknown-linux-gnu",
        ("armv7l", "linux"): f"deep-filter-{DEEP_FILTER_VERSION}-armv7-unknown-linux-gnueabihf",
        ("arm", "linux"): f"deep-filter-{DEEP_FILTER_VERSION}-armv7-unknown-linux-gnueabihf",
        ("x86_64", "darwin"): f"deep-filter-{DEEP_FILTER_VERSION}-x86_64-apple-darwin",
        ("x86_64", "linux"): f"deep-filter-{DEEP_FILTER_VERSION}-x86_64-unknown-linux-musl",
        ("x86_64", "windows"): f"deep-filter-{DEEP_FILTER_VERSION}-x86_64-pc-windows-msvc.exe",
    }

    key = (ARCH, OS)
    if key not in ARCH_OS_MAP:
        print(f"Unsupported architecture or platform: {ARCH}-{OS}")
        sys.exit(1)

    DEEP_FILTER_BINARY = ARCH_OS_MAP[key]
    print(f"Downloading {DEEP_FILTER_BINARY}...")

    download_url = f"{DEEP_FILTER_BASE_URL}/{DEEP_FILTER_BINARY}"
    binary_path = TOOLS_DIR / DEEP_FILTER_BINARY

    try:
        urllib.request.urlretrieve(download_url, binary_path)
    except Exception as e:
        print(f"Error downloading file: {e}")
        sys.exit(1)

    # Rename the binary to "deep-filter" and make it executable
    binary_path.rename(DEEP_FILTER_OUTPUT)
    DEEP_FILTER_OUTPUT.chmod(0o755)
    print(f"DeepFilterNet binary installed as: {DEEP_FILTER_OUTPUT}")


if __name__ == "__main__":
    install_deep_filter()



import concurrent.futures
import os
import shutil
from pathlib import Path
from typing import Any

from waddle.audios.align_offset import align_speaker_to_reference
from waddle.audios.call_tools import (
    convert_all_files_to_wav,
    convert_to_wav,
    remove_noise,
)
from waddle.audios.clip import clip_audio
from waddle.config import DEFAULT_COMP_AUDIO_DURATION, DEFAULT_LANGUAGE
from waddle.processing.combine import (
    combine_audio_files,
    combine_srt_files,
    merge_timelines,
)
from waddle.processing.segment import (
    SpeechTimeline,
    detect_speech_timeline,
    process_segments,
)
from waddle.utils import to_path


def select_reference_audio(audio_paths: list[Path]) -> Path:
    """
    Automatically select a reference audio file starting with 'GMT'.

    Args:
        audio_paths (list): List of audio file paths.

    Returns:
        str: Path to the reference audio file.
    """
    gmt_files = [f for f in audio_paths if Path(f).name.startswith("GMT")]
    if not gmt_files:
        raise ValueError("No reference audio file found and no GMT file exists.")
    return gmt_files[0]


def process_single_file(
    aligned_audio: str | bytes | os.PathLike[Any],
    output_dir: str | bytes | os.PathLike[Any],
    speaker_audio: str | bytes | os.PathLike[Any],
    ss: float = 0.0,
    out_duration: float | None = None,
    no_noise_remove: bool = False,
    whisper_options: str = f"-l {DEFAULT_LANGUAGE}",
) -> Path:
    """
    Process a single audio file: normalize, detect speech, and transcribe.

    Args:
        aligned_audio (str | os.PathLike): Path to the aligned audio file.
        output_dir (str | os.PathLike): Path to the output directory.
        speaker_audio (str | os.PathLike): Path to the speaker audio file.
        ss (float, optional): Start time offset in seconds. Defaults to 0.0.
        out_duration (float | None, optional): Duration of the processed output audio in seconds.

    Returns:
        Path: Path to the combined speaker audio file.
    """
    aligned_audio_path = to_path(aligned_audio)
    output_dir_path = to_path(output_dir)
    speaker_audio_path = to_path(speaker_audio)

    if aligned_audio_path.suffix != ".wav":
        convert_to_wav(aligned_audio_path)
        aligned_audio_path = aligned_audio_path.with_suffix(".wav")
    if ss > 0 or out_duration:
        clip_audio(aligned_audio_path, aligned_audio_path, ss=ss, out_duration=out_duration)
    if not no_noise_remove:
        remove_noise(aligned_audio_path, aligned_audio_path)

    segs_folder_path, _ = detect_speech_timeline(aligned_audio_path)

    # Transcribe segments and combine
    speaker_name = speaker_audio_path.stem
    combined_speaker_path = output_dir_path / f"{speaker_name}.wav"
    transcription_path = output_dir_path / f"{speaker_name}.srt"
    process_segments(
        segs_folder_path,
        combined_speaker_path,
        transcription_path=transcription_path,
        whisper_options=whisper_options,
    )

    return combined_speaker_path


def preprocess_multi_files(
    reference: str | bytes | os.PathLike[Any] | None,
    source_dir: str | bytes | os.PathLike[Any],
    output_dir: str | bytes | os.PathLike[Any],
    comp_duration: float = DEFAULT_COMP_AUDIO_DURATION,
    ss: float = 0.0,
    out_duration: float | None = None,
    no_noise_remove: bool = False,
    convert: bool = True,
    transcribe: bool = False,
    whisper_options: str = f"-l {DEFAULT_LANGUAGE}",
) -> None:
    source_dir_path = to_path(source_dir)
    output_dir_path = to_path(output_dir)

    if output_dir_path.exists():
        shutil.rmtree(output_dir_path, ignore_errors=True)
    output_dir_path.mkdir(parents=True, exist_ok=True)

    # Workspace for temporary files
    workspace_path = source_dir_path / "workspace"
    if workspace_path.exists():
        shutil.rmtree(workspace_path, ignore_errors=True)
    workspace_path.mkdir(parents=True, exist_ok=True)

    # Convert to WAV files if the flag is set
    if convert:
        print("[INFO] Converting audio files to WAV format...")
        convert_all_files_to_wav(source_dir_path)

    audio_file_paths = sorted(source_dir_path.glob("*.wav"))
    if not audio_file_paths:
        raise ValueError("No audio files found in the directory.")

    reference_path = to_path(reference) if reference else select_reference_audio(audio_file_paths)
    print(f"[INFO] Using reference audio: {reference_path}")

    audio_file_paths = [f for f in audio_file_paths if f != reference_path and "GMT" not in f.name]
    if not audio_file_paths:
        raise ValueError("No speaker audio files found in the directory.")

    timelines: list[SpeechTimeline] = []
    segments_dir_list = []

    def process_file(speaker_audio_path: Path):
        print(f"\033[92m[INFO] Processing file: {str(speaker_audio_path)}\033[0m")

        # 1) Align each speaker audio to the reference
        aligned_audio_path = align_speaker_to_reference(
            reference_path,
            speaker_audio_path,
            workspace_path,
            comp_duration=comp_duration,
        )
        clip_audio(aligned_audio_path, aligned_audio_path, ss=ss, out_duration=out_duration)
        if not no_noise_remove:
            remove_noise(aligned_audio_path, aligned_audio_path)

        # 2) Preprocess the aligned audio file
        segments_dir, timeline = detect_speech_timeline(aligned_audio_path)

        return segments_dir, timeline

    with concurrent.futures.ThreadPoolExecutor() as executor:
        results = list(executor.map(process_file, audio_file_paths))

    for segments_dir, timeline in results:
        segments_dir_list.append(segments_dir)
        timelines.append(timeline)

    merged_timeline = merge_timelines(timelines)

    for audio_file_path, segments_dir in zip(audio_file_paths, segments_dir_list, strict=False):
        speaker_name = audio_file_path.stem
        combined_speaker_path = output_dir_path / f"{speaker_name}.wav"
        # If transcribe is False, transcription_path is None and not transcribed.
        transcription_path = output_dir_path / f"{speaker_name}.srt" if transcribe else None
        process_segments(
            segments_dir,
            combined_speaker_path,
            transcription_path=transcription_path,
            whisper_options=whisper_options,
            timeline=merged_timeline,
        )

    # Clean up workspace_path
    shutil.rmtree(workspace_path, ignore_errors=True)

    if not transcribe:
        return

    audio_prefix = audio_file_paths[0].stem
    if "-" in audio_prefix:
        audio_prefix = audio_prefix.split("-")[0]
    transcription_output_path = output_dir_path / f"{audio_prefix}.srt"
    combine_srt_files(output_dir_path, transcription_output_path)


def postprocess_multi_files(
    source_dir: str | bytes | os.PathLike[Any],
    output_dir: str | bytes | os.PathLike[Any],
    ss: float = 0.0,
    out_duration: float | None = None,
    whisper_options: str = f"-l {DEFAULT_LANGUAGE}",
) -> None:
    source_dir_path = to_path(source_dir)
    output_dir_path = to_path(output_dir)

    audio_file_paths = [f for f in sorted(source_dir_path.glob("*.wav")) if "GMT" not in f.name]
    if not audio_file_paths:
        raise ValueError("No audio files found in the directory.")

    if output_dir_path.exists():
        shutil.rmtree(output_dir_path, ignore_errors=True)
    output_dir_path.mkdir(parents=True, exist_ok=True)

    for audio_file_path in audio_file_paths:
        tmp_audio_file_path = output_dir_path / audio_file_path.name
        if ss > 0 or out_duration:
            clip_audio(audio_file_path, tmp_audio_file_path, ss=ss, out_duration=out_duration)
        else:
            shutil.copy(audio_file_path, tmp_audio_file_path)
        segments_dir, _ = detect_speech_timeline(tmp_audio_file_path)
        speaker_name = audio_file_path.stem
        combined_speaker_path = output_dir_path / f"{speaker_name}.wav"
        transcription_path = output_dir_path / f"{speaker_name}.srt"
        process_segments(
            segments_dir,
            combined_speaker_path,
            transcription_path=transcription_path,
            whisper_options=whisper_options,
        )

    audio_prefix = audio_file_paths[0].stem
    if "-" in audio_prefix:
        audio_prefix = audio_prefix.split("-")[0]

    transcription_output_path = output_dir_path / f"{audio_prefix}.srt"
    combine_srt_files(output_dir_path, transcription_output_path)
    shutil.copy(transcription_output_path, output_dir_path / f"{audio_prefix}.md")

    final_audio_path = output_dir_path / f"{audio_prefix}.wav"
    combined_audio_paths = sorted(output_dir_path.glob("*.wav"))
    combine_audio_files(combined_audio_paths, final_audio_path)



DEFAULT_LANGUAGE = "ja"

DEFAULT_OUT_AUDIO_DURATION = 60.0 * 1  # seconds

DEFAULT_COMP_AUDIO_DURATION = 60.0 * 20  # seconds
DEFAULT_SR = 48000  # Hz

DEFAULT_CHUNK_DURATION = 0.5  # seconds
DEFAULT_BUFFER_DURATION = 0.25  # seconds

DEFAULT_TARGET_DB = -30.0  # dBFS
DEFAULT_THRESHOLD_DB = -70.0  # dBFS

APP_NAME = "waddle"
APP_AUTHOR = "emptymap"



import pytest

from waddle.argparse import create_waddle_parser


def test_ss_0():
    parser = create_waddle_parser()
    args = parser.parse_args(["single", "audio", "-ss", "00:00:10,050"])
    assert args.ss == 10.05


def test_ss_1():
    parser = create_waddle_parser()
    args = parser.parse_args(["single", "audio", "-ss", "30.5"])
    assert args.ss == 30.5


def test_ss_2():
    parser = create_waddle_parser()
    args = parser.parse_args(["single", "audio", "-ss", "99:59:59,999"])
    assert args.ss == 359999.999


def test_ss_3():
    parser = create_waddle_parser()
    args = parser.parse_args(["single", "audio", "-ss", "0"])
    assert args.ss == 0.0


def test_ss_4():
    parser = create_waddle_parser()
    args = parser.parse_args(["single", "audio", "-ss", "0.0"])
    assert args.ss == 0.0


def test_ss_5():
    parser = create_waddle_parser()
    with pytest.raises(SystemExit):
        parser.parse_args(["single", "audio", "-ss", "00:00:00:00"])


def test_ss_6():
    parser = create_waddle_parser()
    with pytest.raises(SystemExit):
        parser.parse_args(["single", "audio", "-ss", "00:00:00.123.53"])



import os
import subprocess
from pathlib import Path

from platformdirs import user_runtime_dir

from waddle.config import APP_AUTHOR, APP_NAME


def install_whisper_cpp():
    # Tool installation directories
    TOOLS_DIR = Path(user_runtime_dir(APP_NAME, APP_AUTHOR)) / "tools"
    WHISPER_DIR = TOOLS_DIR / "whisper.cpp"

    # Create the tools directory if it doesn't exist
    TOOLS_DIR.mkdir(parents=True, exist_ok=True)

    # Clone whisper.cpp if not already cloned
    if not WHISPER_DIR.is_dir():
        print("Cloning whisper.cpp repository...")
        subprocess.run(
            ["git", "clone", "https://github.com/ggerganov/whisper.cpp.git", str(WHISPER_DIR)],
            check=True,
        )
        subprocess.run(["git", "checkout", "v1.7.4"], check=True, cwd=str(WHISPER_DIR))
    else:
        print(f"whisper.cpp already exists at {WHISPER_DIR}")

    # Check if WHISPER_MODEL_NAME is defined, if not assign "large-v3" as default
    WHISPER_MODEL_NAME = os.getenv("WHISPER_MODEL_NAME", "large-v3")
    print(f"WHISPER_MODEL_NAME is set to: {WHISPER_MODEL_NAME}")

    # Download the model if not already downloaded
    model_path = WHISPER_DIR / "models" / f"ggml-{WHISPER_MODEL_NAME}.bin"
    if not model_path.is_file():
        print(f"Downloading the {WHISPER_MODEL_NAME} model...")
        subprocess.run(
            ["sh", "./models/download-ggml-model.sh", WHISPER_MODEL_NAME],
            check=True,
            cwd=WHISPER_DIR,
        )
    else:
        print(f"{WHISPER_MODEL_NAME} model already exists.")

    # Build the project
    print("Building whisper.cpp...")
    subprocess.run(["cmake", "-B", "build"], check=True, cwd=str(WHISPER_DIR))
    subprocess.run(
        ["cmake", "--build", "build", "--config", "Release"], check=True, cwd=str(WHISPER_DIR)
    )

    print(f"whisper.cpp installed successfully in: {WHISPER_DIR}")


if __name__ == "__main__":
    install_whisper_cpp()



import os
import subprocess
import tempfile
import threading
from pathlib import Path

from platformdirs import user_runtime_dir

from waddle.config import APP_AUTHOR, APP_NAME, DEFAULT_LANGUAGE
from waddle.tools.install_deep_filter import install_deep_filter
from waddle.tools.install_whisper_cpp import install_whisper_cpp


def get_tools_dir() -> Path:
    """
    Get the tools directory.
    """
    return Path(user_runtime_dir(APP_NAME, APP_AUTHOR)) / "tools"


def convert_to_mp3(
    input_path: Path, output_path_or_none: Path | None = None, force: bool = False
) -> None:
    """Convert audio file to MP3 format."""
    output_path = output_path_or_none or input_path.with_suffix(".mp3")
    if output_path.exists() and not force:
        print(f"[INFO] Skipping {input_path}: MP3 file already exists.")
        return

    # Convert to MP3 using ffmpeg
    print(f"[INFO] Converting {input_path} to {output_path}...")
    try:
        subprocess.run(
            [
                "ffmpeg",
                "-y",
                "-i",
                str(input_path),
                "-q:a",
                "0",
                str(output_path),
            ],
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        print(f"[INFO] Successfully converted: {output_path}")
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"[ERROR] Converting {input_path}: {e}") from e


def convert_to_wav(input_path: Path, output_path_or_none: Path | None = None) -> None:
    """Convert audio file to WAV format."""
    output_path = output_path_or_none or input_path.with_suffix(".wav")
    if output_path.exists():
        print(f"[INFO] Skipping {input_path}: WAV file already exists.")
        return

    # Convert to WAV using ffmpeg
    print(f"[INFO] Converting {input_path} to {output_path}...")
    try:
        subprocess.run(
            [
                "ffmpeg",
                "-y",
                "-i",
                str(input_path),
                str(output_path),
            ],  # Added '-y' flag
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        print(f"[INFO] Successfully converted: {output_path}")
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"[ERROR] Converting {input_path}: {e}") from e


def convert_all_files_to_wav(folder_path: Path) -> None:
    """Convert all audio files in the specified folder to WAV format."""
    # File extensions to look for
    valid_extensions = (".m4a", ".aifc", ".mp4")

    # Find all valid files in the current directory and subdirectories
    folder = Path(folder_path)
    for input_path in folder.rglob("*"):
        if input_path.suffix in valid_extensions:
            convert_to_wav(input_path)


def ensure_sampling_rate(
    input_path: Path, output_path: Path, target_rate: int, bit_depth: str = "16"
) -> None:
    """
    Ensure the input WAV file has the specified sampling rate and bit depth.
    Converts the input file if needed.
    """
    if not input_path.exists():
        raise FileNotFoundError(f"Input file not found: {input_path}")

    command = [
        "ffmpeg",
        "-i",
        str(input_path),  # Input file
        "-ar",
        str(target_rate),  # Set target sample rate
        "-ac",
        "1",  # Ensure mono channel
        "-c:a",
        f"pcm_s{bit_depth}le",  # Set bit depth
        str(output_path),  # Output file
        "-y",  # Overwrite output file
    ]

    try:
        subprocess.run(
            command,
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"[ERROR] Converting {input_path} to {output_path}: {e}") from e


deep_filter_install_lock = threading.Lock()


def remove_noise(input_path: Path, output_path: Path) -> None:
    """
    Enhance audio by removing noise using DeepFilterNet.
    """
    # Paths
    tools_dir = get_tools_dir()
    deep_filter_path = tools_dir / "deep-filter"
    tmp_file_path = input_path.with_stem(input_path.stem + "_tmp")

    # Check if the tool exists
    with deep_filter_install_lock:
        if not deep_filter_path.exists():
            install_deep_filter()

    # Ensure input is 48kHz and 16-bit
    ensure_sampling_rate(input_path, tmp_file_path, target_rate=48000)

    # Run the DeepFilterNet tool without printing its output
    output_folder_path = output_path.parent
    command = [str(deep_filter_path), str(tmp_file_path), "-o", str(output_folder_path)]
    try:
        subprocess.run(command, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        output_path.write_bytes(tmp_file_path.read_bytes())

    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"[ERROR] Running DeepFilterNet: {e}") from e

    finally:
        # Cleanup temporary file
        if tmp_file_path.exists():
            tmp_file_path.unlink()


whisper_install_lock = threading.Lock()


def transcribe_in_batches(
    input_output_paths: list[tuple[Path, Path]],
    options: str = f"-l {DEFAULT_LANGUAGE}",
    *,
    batch_size: int = 200,
) -> None:
    """
    Transcribe audio in batches using Whisper.cpp.
    """

    # Paths
    tools_dir = get_tools_dir()
    whisper_bin = tools_dir / "whisper.cpp" / "build" / "bin" / "whisper-cli"
    whisper_model = (
        tools_dir
        / "whisper.cpp"
        / "models"
        / f"ggml-{os.getenv('WHISPER_MODEL_NAME') or 'large-v3'}.bin"
    )

    # Check if the Whisper binary exists
    with whisper_install_lock:
        if not whisper_model.exists():
            install_whisper_cpp()

    # Check if the Whisper model exists
    if not whisper_model.exists():
        raise FileNotFoundError(f"Whisper model not found. Please ensure {whisper_model} exists.")

    # List of pairs of tmp audio paths and output paths
    tmp_output_paths = []
    with tempfile.TemporaryDirectory() as temp_dir:
        for input_path, output_path in input_output_paths:
            # Ensure input is 16kHz and 16-bit
            tmp_audio_path = Path(temp_dir) / input_path.name
            ensure_sampling_rate(input_path, tmp_audio_path, target_rate=16000)
            tmp_output_paths.append((tmp_audio_path, output_path))

        batches = [
            tmp_output_paths[i : i + batch_size]
            for i in range(0, len(tmp_output_paths), batch_size)
        ]

        for batch in batches:
            command = [
                str(whisper_bin),
                "-m",
                str(whisper_model),
                *options.split(),
                "-osrt",
            ]
            for temp_audio_path, output_path in batch:
                command.extend([str(temp_audio_path), "-of", str(output_path)])
            try:
                subprocess.run(
                    command, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
                )
                for _, output_path in batch:
                    Path(f"{output_path}.srt").replace(output_path)
            except subprocess.CalledProcessError as e:
                raise RuntimeError(f"[ERROR] Running Whisper: {e}") from e


def transcribe(
    input_path: Path, output_path: Path, options: str = f"-l {DEFAULT_LANGUAGE}"
) -> None:
    """
    Transcribe audio using Whisper.cpp.
    """
    # Paths
    tools_dir = get_tools_dir()
    whisper_bin = tools_dir / "whisper.cpp" / "build" / "bin" / "whisper-cli"
    whisper_model = (
        tools_dir
        / "whisper.cpp"
        / "models"
        / f"ggml-{os.getenv('WHISPER_MODEL_NAME') or 'large-v3'}.bin"
    )
    temp_audio_path = input_path.with_stem(input_path.stem + "_16k_16bit")

    # Check if the Whisper binary exists
    with whisper_install_lock:
        if not whisper_model.exists():
            install_whisper_cpp()

    # Check if the Whisper model exists
    if not whisper_model.exists():
        raise FileNotFoundError(f"Whisper model not found. Please ensure {whisper_model} exists.")

    # Ensure input is 16kHz and 16-bit
    ensure_sampling_rate(input_path, temp_audio_path, target_rate=16000)

    # Transcribe using Whisper without printing its output
    command = [
        str(whisper_bin),
        "-m",
        str(whisper_model),
        "-f",
        str(temp_audio_path),
        *options.split(),
        "-osrt",
        "-of",
        output_path,
    ]
    try:
        subprocess.run(command, check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        Path(f"{output_path}.srt").replace(output_path)
    except subprocess.CalledProcessError as e:
        raise RuntimeError(f"[ERROR] Running Whisper: {e}") from e
    finally:
        # Clean up the temporary file
        if temp_audio_path.exists():
            temp_audio_path.unlink()



from waddle.processing.segment import merge_segments


def test_merge_segments_01():
    segments = [(0, 100), (200, 300)]
    assert merge_segments(segments) == [(0, 100), (200, 300)]


def test_merge_segments_02():
    segments = [(0, 100), (100, 200), (200, 300)]
    assert merge_segments(segments) == [(0, 300)]


def test_merge_segments_03():
    segments = [(0, 300), (200, 400), (500, 600)]
    assert merge_segments(segments) == [(0, 400), (500, 600)]


def test_merge_segments_04():
    segments = []
    assert merge_segments(segments) == []



import shutil
from pathlib import Path
from typing import Optional, TypeAlias

from pydub import AudioSegment

from waddle.utils import parse_audio_filename

SpeechSegment: TypeAlias = tuple[int, int]
SpeechTimeline: TypeAlias = list[SpeechSegment]

SrtEntry: TypeAlias = tuple[str, str, str]
SrtEntries: TypeAlias = list[SrtEntry]


def combine_segments_into_audio(
    segs_folder_path: Path,
    combined_audio_path: Path,
    timeline: SpeechTimeline | None = None,
) -> None:
    segs_folder_path = Path(segs_folder_path)  # TODO: Delete it after switch to Pathlib in test
    combined_audio_path = Path(
        combined_audio_path
    )  # TODO: Delete it after switch to Pathlib in test

    seg_file_paths = sorted(segs_folder_path.glob("*.wav"), key=lambda x: int(x.stem.split("_")[1]))
    max_end_ms = adjust_pos_to_timeline(timeline, timeline[-1][1]) if timeline else 1
    if not seg_file_paths:
        print("\033[93m[WARNING] No segment files found for combining.\033[0m")

        # Output a dummy audio file
        final_audio = AudioSegment.silent(duration=max_end_ms)
        final_audio.export(combined_audio_path, format="wav")

        # Clean up segs folder
        shutil.rmtree(segs_folder_path, ignore_errors=True)
        return

    if not timeline:
        _, end_ms = parse_audio_filename(str(seg_file_paths[-1]))
        timeline = [(0, end_ms)]
        max_end_ms = end_ms

    final_audio = AudioSegment.silent(duration=max_end_ms)

    for seg_file_path in seg_file_paths:
        seg_audio = AudioSegment.from_file(str(seg_file_path))
        start_ms, _ = parse_audio_filename(str(seg_file_path))
        final_audio = final_audio.overlay(
            seg_audio, position=adjust_pos_to_timeline(timeline, start_ms)
        )

    final_audio.export(str(combined_audio_path), format="wav")


def adjust_pos_to_timeline(timeline: SpeechTimeline, pos: int) -> int:
    """
    Adjust the position in the original audio to an edited audio.
    This function converts a position in the original audio to the corresponding position
    in the edited audio, based on the provided timeline which specifies the valid audio
    segments and omits silence.

    Args:
        timeline (SpeechTimeline): A list of tuples where each tuple contains the start
                                   and end positions of valid audio segments.
        pos (int): The position in the original audio to be adjusted.
    Returns:
        int: The adjusted position in the edited audio.
    """

    running_total = 0
    for start, end in timeline:
        if pos <= start:
            # if `start_ms` is before this segment, it doesn't matter
            break
        if start < pos <= end:
            # `start_ms` is within this segment.
            # Add the valid time in this segment to `running_total` and return
            return running_total + (pos - start)
        # end < start_ms must hold
        # Add the length of this segment to `running_total`
        running_total += end - start
    return running_total


def combine_audio_files(aligned_audio_paths: list[Path], output_audio_path: Path) -> None:
    """
    Combine multiple aligned audio files by overlaying them sequentially.

    Args:
        aligned_audio_paths (list): Paths to aligned audio files.
        output_audio_path (str): Path to save the combined audio file.
    """
    combined_audio = None
    # Sort for making max duration audio
    aligned_audio_paths.sort(key=lambda x: AudioSegment.from_file(x).duration_seconds, reverse=True)
    for path in aligned_audio_paths:
        audio = AudioSegment.from_file(str(path))
        combined_audio = audio if combined_audio is None else combined_audio.overlay(audio)

    if combined_audio:
        combined_audio.export(output_audio_path, format="wav")


def parse_srt(file_path: Path, speaker_name: Optional[str] = None) -> SrtEntries:
    """
    Parse an SRT file, attach speaker name to each text line,
    and return a list of (start_time_for_sorting, end_timestamp, text_with_speaker).

    Args:
        file_path (str): Path to the .srt file.
        speaker_name (str): Name of the speaker.

    Returns:
        list: List of tuples (start_time_for_sorting, end_timestamp, text_with_speaker).
    """
    entries: SrtEntries = []
    with open(str(file_path), "r", encoding="utf-8") as f:
        blocks = f.read().strip().split("\n\n")

    for block in blocks:
        lines = block.split("\n")
        if len(lines) < 3:
            continue

        timestamp = lines[1]
        start, end = timestamp.split(" --> ")
        text = " ".join(lines[2:])
        if speaker_name:
            text = f"{speaker_name}: {text}"

        # Convert timestamps for sorting
        start_time_for_sorting = start.replace(",", ".")
        end_time_for_sorting = end.replace(",", ".")
        entries.append((start_time_for_sorting, end_time_for_sorting, text))
    return entries


def combine_srt_files(input_dir_path: Path, output_file_path: Path) -> None:
    """
    Combine all .srt files in a directory, add speaker names, and sort by timestamp.
    """
    input_dir_path = Path(input_dir_path)  # TODO: Delete it after switch to Pathlib in test
    output_file_path = Path(output_file_path)  # TODO: Delete it after switch to Pathlib in test
    srt_file_paths = [f for f in input_dir_path.glob("*.srt")]
    all_entries: SrtEntries = []

    for srt_file_path in srt_file_paths:
        srt_file_name = srt_file_path.stem
        if "-" in srt_file_name:
            srt_file_name = srt_file_name.split("-")[1]
        all_entries.extend(parse_srt(srt_file_path, srt_file_name))
        srt_file_path.unlink()

    # Sort all entries by timestamp
    all_entries.sort(key=lambda x: x[0])

    # Write combined SRT file
    with open(str(output_file_path), "w", encoding="utf-8") as f:
        for i, (start, end, text) in enumerate(all_entries, start=1):
            f.write(f"{i}\n")
            f.write(f"{start} --> {end}\n")
            f.write(f"{text}\n\n")


def merge_timelines(timelines: list[SpeechTimeline]) -> SpeechTimeline:
    """
    Merges multiple speech timelines into a single, non-overlapping timeline.
    Args:
        timelines (list[SpeechTimeline]): A list of SpeechTimeline objects,
                                          where each SpeechTimeline is a list of tuples.
                                          Each tuple represents a segment with a start and end time.
    Returns:
        SpeechTimeline: A merged SpeechTimeline with non-overlapping segments,
                        sorted by start time.
    """

    all_segments: SpeechTimeline = []
    for timeline in timelines:
        all_segments.extend(timeline)

    # Sort and merge overlapping segments
    all_segments = list(set(all_segments))
    all_segments.sort(key=lambda x: x[0])

    merged: SpeechTimeline = []
    for seg in all_segments:
        # If merged is empty or current segment does not overlap the last one in merged
        if not merged or merged[-1][1] < seg[0]:
            merged.append(seg)
        else:
            merged[-1] = (merged[-1][0], max(merged[-1][1], seg[1]))

    return merged



import os
from pathlib import Path


def time_to_seconds(timestamp: str) -> float:
    """
    Convert SRT timestamp format (hh:mm:ss,ms) to seconds.

    Args:
        timestamp (str): Timestamp in the format "hh:mm:ss,ms".

    Returns:
        float: Time in seconds.
    """
    try:
        hours, minutes, seconds = timestamp.replace(",", ".").split(":")
        return int(hours) * 3600 + int(minutes) * 60 + float(seconds)
    except Exception as e:
        raise ValueError(f"Invalid timestamp format: {timestamp}") from e


def phrase_time_to_seconds(s: str) -> float:
    try:
        return float(s)
    except ValueError:
        pass

    value = s.replace(",", ".")
    parts = list(map(float, value.split(":")))

    if len(parts) > 3:
        raise ValueError(f"Invalid time format: {s}")
    weights = [3600, 60, 1][-len(parts) :]

    return float(sum(p * w for p, w in zip(parts, weights, strict=False)))


def format_time(seconds: float) -> str:
    """
    Format seconds to SRT timestamp format (hh:mm:ss,ms).

    Args:
        seconds (float): Time in seconds.

    Returns:
        str: Timestamp in SRT format "hh:mm:ss,ms".
    """
    hh = int(seconds // 3600)
    mm = int((seconds % 3600) // 60)
    ss = int(seconds % 60)
    ms = round((seconds % 1) * 1000)
    return f"{hh:02}:{mm:02}:{ss:02},{ms:03}"


def format_audio_filename(prefix: str, start: int, end: int) -> str:
    """Generate a standardized audio filename with a given prefix and time range."""
    return f"{prefix}_{start}_{end}.wav"


def parse_audio_filename(filename: str) -> tuple:
    """Extract and return the start and end timestamps from a standardized audio filename."""
    parts = filename.split("_")
    start_str, end_str = parts[-2], parts[-1].split(".")[0]
    return int(start_str), int(end_str)


def to_path(obj: str | bytes | os.PathLike) -> Path:
    """Converts input to a pathlib.Path object, handling str, bytes, and os.PathLike."""
    if isinstance(obj, Path):
        return obj

    fs_path = os.fspath(obj)
    if isinstance(fs_path, (bytes, bytearray, memoryview)):
        fs_path = bytes(fs_path).decode()
    return Path(fs_path)



import os
from pathlib import Path

import pytest

from waddle.utils import (
    format_audio_filename,
    format_time,
    parse_audio_filename,
    phrase_time_to_seconds,
    time_to_seconds,
    to_path,
)


def test_time_to_seconds():
    """Test conversion of SRT timestamps to seconds."""
    assert time_to_seconds("00:00:00,000") == 0.0, "time_to_seconds('00:00:00,000') is failed."
    assert time_to_seconds("00:01:30,500") == 90.5, "time_to_seconds('00:01:30,500') is failed."
    assert time_to_seconds("01:23:45,678") == 5025.678, "time_to_seconds('01:23:45,678') is failed."
    assert time_to_seconds("12:34:56,789") == 45296.789, (
        "time_to_seconds('12:34:56,789') is failed."
    )
    assert time_to_seconds("99:59:59,999") == 359999.999, (
        "time_to_seconds('99:59:59,999') is failed."
    )

    with pytest.raises(ValueError):
        time_to_seconds("00:00:00:00")
    with pytest.raises(ValueError):
        time_to_seconds("00:00:00.123.53")
    with pytest.raises(ValueError):
        time_to_seconds("00:00.123")
    with pytest.raises(ValueError):
        time_to_seconds("00.12")


def test_phrase_time_to_seconds():
    """Test conversion of argument timestamps to seconds."""
    assert phrase_time_to_seconds("00:01") == 1.0
    assert phrase_time_to_seconds("00:01.5") == 1.5
    assert phrase_time_to_seconds("05:12") == 312.0
    assert phrase_time_to_seconds("05:12.5") == 312.5
    assert phrase_time_to_seconds("12") == 12.0
    assert phrase_time_to_seconds("12.5") == 12.5
    assert phrase_time_to_seconds("12:34:56") == 45296.0
    assert phrase_time_to_seconds("12:34:56.789") == 45296.789
    assert phrase_time_to_seconds("00:1.5:00") == 90.0
    assert phrase_time_to_seconds("00:1.5:10") == 100.0
    assert phrase_time_to_seconds("0.1:0.2:0.3") == 372.3


def test_format_time():
    """Test formatting of seconds into SRT timestamp format."""
    assert format_time(0.0) == "00:00:00,000", "format_time(0.0) is failed."
    assert format_time(90.5) == "00:01:30,500", "format_time(90.5) is failed."
    assert format_time(5025.678) == "01:23:45,678", "format_time(5025.678) is failed."
    assert format_time(45296.789) == "12:34:56,789", "format_time(45296.789) is failed."
    assert format_time(359999.999) == "99:59:59,999", "format_time(359999.999) is failed."


def test_format_audio_filename():
    """Test get segs/chunks name."""
    assert format_audio_filename("chunk", 0, 100) == "chunk_0_100.wav", (
        "get_name('chunk', 0, 100) is failed."
    )
    assert format_audio_filename("chunk", 1, 200) == "chunk_1_200.wav", (
        "get_name('chunk', 1, 200) is failed."
    )
    assert format_audio_filename("seg", 1, 10) == "seg_1_10.wav", (
        "get_name('seg', 1, 10) is failed."
    )
    assert format_audio_filename("seg", 10, 20) == "seg_10_20.wav", (
        "get_name('seg', 10, 20) is failed."
    )


def test_parse_audio_filename():
    """Test parse segs/chunks to get start and end."""
    assert parse_audio_filename("chunk_0_100.wav") == (0, 100), (
        "parse_name('chunk_0_100.wav') is failed."
    )
    assert parse_audio_filename("chunk_1_200.wav") == (1, 200), (
        "parse_name('chunk_1_200.wav') is failed."
    )
    assert parse_audio_filename("seg_1_10.wav") == (1, 10), "parse_name('seg_1_10.wav') is failed."
    assert parse_audio_filename("seg_10_20.wav") == (10, 20), (
        "parse_name('seg_10_20.wav') is failed."
    )

    # Other folder
    assert parse_audio_filename("folder/chunk_0_100.wav") == (0, 100), (
        "parse_name('folder/chunk_0_100.wav') is failed."
    )
    assert parse_audio_filename("../folder/seg_1_200.wav") == (1, 200), (
        "parse_name('../folder/seg_1_200.wav') is failed."
    )

    # _ is used in the folder name
    assert parse_audio_filename("tmp/kzjirwe_klae256_wj1/seg_1_200.wav") == (1, 200), (
        "parse_name('tmp/kzjirwe_klae256_wj1/seg_1_200.wav') is failed."
    )


class PathLikeBytes(os.PathLike):
    """Custom os.PathLike class returning bytes."""

    def __init__(self, path: bytes):
        if not isinstance(path, bytes):
            raise TypeError("Path must be a bytes object")
        self._path = path

    def __fspath__(self) -> bytes:
        return self._path


def test_to_path():
    """Test convert string to Path."""
    assert to_path(Path("abc/23/a.txt")) == Path("abc/23/a.txt")
    assert to_path("test") == Path("test")
    assert to_path("test/test") == Path("test/test")
    assert to_path(os.path.join("test", "test")) == Path("test/test")
    assert to_path(b"z/1/2") == Path("z/1/2")
    assert to_path(b"test/test.png") == Path("test/test.png")
    assert to_path(PathLikeBytes(b"a/b.py")) == Path("a/b.py")



import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from textwrap import dedent

import pytest
from mutagen.id3 import ID3

from waddle.audios.call_tools import convert_to_mp3
from waddle.metadata import (
    ChapterMarker,
    ShowNotesEntry,
    SRTEntry,
    extract_metadata,
    format_time,
    generate_metadata,
    parse_annotated_srt,
)

# Define test directory paths
TESTS_DIR_PATH = Path(__file__).resolve().parents[2] / "tests"
EP0_DIR_PATH = TESTS_DIR_PATH / "ep0"


def test_format_time_01():
    assert format_time(0.0) == "00:00"
    assert format_time(1.0) == "00:01"
    assert format_time(60.0) == "01:00"
    assert format_time(61.0) == "01:01"
    assert format_time(3600.0) == "01:00:00"
    assert format_time(3661.0) == "01:01:01"


def test_parse_annotated_srt_01():
    annotated_srt = (
        dedent("""\
        1
        00:00:00.490 --> 00:00:02.430
        shun: ラバーダックFMへようこそ

        2
        00:00:02.430 --> 00:00:05.970
        shun: このポッドキャストは大学で出会った3人のホストが

        3
        00:00:05.970 --> 00:00:08.650
        shun: 自分たちの興味のあるコンピューターサイエンスや

        4
        00:00:08.650 --> 00:00:10.410
        shun: プログラミングのトピックについて

        5
        00:00:10.410 --> 00:00:13.190
        shun: 自由気ままに語り合うポッドキャストです\n\n""")
    )  # NOTE: SRT-entry must end with two newlines
    result = parse_annotated_srt(annotated_srt)
    assert len(result) == 5
    assert all(isinstance(item, SRTEntry) for item in result)


def test_parse_annotated_srt_02():
    annotated_srt = dedent("""\
        # イントロ
        1
        00:00:00.490 --> 00:00:02.430
        shun: ラバーダックFMへようこそ

        2
        00:00:02.430 --> 00:00:05.970
        shun: このポッドキャストは大学で出会った3人のホストが

        3
        00:00:05.970 --> 00:00:08.650
        shun: 自分たちの興味のあるコンピューターサイエンスや

        4
        00:00:08.650 --> 00:00:10.410
        shun: プログラミングのトピックについて

        5
        00:00:10.410 --> 00:00:13.190
        shun: 自由気ままに語り合うポッドキャストです\n\n""")
    result = parse_annotated_srt(annotated_srt)
    assert len(result) == 6
    assert isinstance(result[0], ChapterMarker)
    assert result[0].title == "イントロ"
    assert result[0].level == 1


def test_parse_annotated_srt_03():
    annotated_srt = dedent("""\
    14
    00:00:36.070 --> 00:00:38.810
    shun: じゃあ甲太郎くんからお願いします

    ## Kotaro: さくらインターネット

    15
    00:00:39.750 --> 00:00:53.870
    kotaro: はいそうですね最近見てたのだと桜インターネットに人がいっぱい座れてますねってことが会社内でも話題になってました

    - [ガバメントクラウドの高すぎるハードル、国産勢唯一さくらインターネットの挑戦 | 日経クロステック（xTECH）](https://xtech.nikkei.com/atcl/nxt/column/18/03074/012300002/))
    """)  # noqa: E501
    result = parse_annotated_srt(annotated_srt)
    assert len(result) == 4
    assert isinstance(result[1], ChapterMarker)
    assert result[1].title == "Kotaro: さくらインターネット"
    assert result[1].level == 2
    assert isinstance(result[3], ShowNotesEntry)
    assert (
        result[3].text
        == "- [ガバメントクラウドの高すぎるハードル、国産勢唯一さくらインターネットの挑戦 | 日経クロステック（xTECH）](https://xtech.nikkei.com/atcl/nxt/column/18/03074/012300002/))"
    )


def test_extract_metadata_01():
    annotated_srt = dedent("""
        # イントロ

        1
        00:00:00.000 --> 00:00:02.000
        shun: ラバーダックFMへようこそ

        # ニュース/コンテンツ

        2
        00:00:02.000 --> 00:00:05.000
        shun: では、各々が最近興味を持ったニュースやコンテンツを紹介していきます

        ## Kotaro: さくらインターネット

        3
        00:00:05.000 --> 00:00:10.000
        kotaro: 桜インターネットに人が吸われてますね

        - [ガバメントクラウドの高すぎるハードル、国産勢唯一さくらインターネットの挑戦 | 日経クロステック（xTECH）](https://xtech.nikkei.com/atcl/nxt/column/18/03074/012300002/)

        ## Masa: LLM

        4
        00:00:10.000 --> 00:00:15.000
        masa: 最近AI関連、LLM関連で読んで面白かった論文がありました。

        - [[2307.03172] Lost in the Middle: How Language Models Use Long Contexts](https://arxiv.org/abs/2307.03172)
        - [[2406.16008] Found in the Middle: Calibrating Positional Attention Bias Improves Long Context Utilization](https://arxiv.org/abs/2406.16008)

        # アウトロ

        5
        00:00:15.000 --> 00:00:20.000
        shun: それでは、今回はこの辺で\n\n""")  # noqa: E501
    entries = parse_annotated_srt(annotated_srt)
    chapters, show_notes = extract_metadata(entries)
    assert len(chapters) == 5
    assert chapters[0].title == "イントロ"
    assert chapters[0].start == 0.0
    assert chapters[0].end == 2.0
    assert chapters[1].title == "ニュース/コンテンツ"
    assert chapters[2].title == "Kotaro: さくらインターネット"
    assert chapters[3].title == "Masa: LLM"
    assert chapters[4].title == "アウトロ"

    assert (
        show_notes.strip()
        == dedent("""
        - [ガバメントクラウドの高すぎるハードル、国産勢唯一さくらインターネットの挑戦 | 日経クロステック（xTECH）](https://xtech.nikkei.com/atcl/nxt/column/18/03074/012300002/)
        - [[2307.03172] Lost in the Middle: How Language Models Use Long Contexts](https://arxiv.org/abs/2307.03172)
        - [[2406.16008] Found in the Middle: Calibrating Positional Attention Bias Improves Long Context Utilization](https://arxiv.org/abs/2406.16008)\n
        """).strip()  # noqa: E501
    )


def test_extract_metadata_02():
    annotated_srt = dedent("""
        # Chapter with no srt entries
        ## ↑ gets ignored and this will be the first chapter

        1
        00:00:00.000 --> 00:00:02.000
        alice: Hello

        # This also will be ignored
    """)
    entries = parse_annotated_srt(annotated_srt)
    chapters, show_notes = extract_metadata(entries)
    assert len(chapters) == 1
    assert chapters[0].title.startswith("↑ gets ignored")
    assert show_notes == ""


def test_extract_metadata_03():
    annotated_srt = dedent("""
        1
        00:00:00.000 --> 00:00:02.000
        alice: Hello

        Can write Markdown here.

        - List 1
        - List 2
          - Sublist 1

        2
        00:00:02.000 --> 00:00:04.000
        bob: Hi

        - The list continues here

        ;
        > `;` can be used to insert newline
        > empty lines are ignored otherwise
        ;
        ; Can also write something here
        ;

        3
        00:00:04.000 --> 00:00:06.000
        alice: How are you?\n\n

        The last line will be included in the show notes.""")
    entries = parse_annotated_srt(annotated_srt)
    chapters, show_notes = extract_metadata(entries)
    assert len(chapters) == 0
    assert (
        show_notes.strip()
        == dedent("""
        Can write Markdown here.
        - List 1
        - List 2
          - Sublist 1
        - The list continues here

        > `;` can be used to insert newline
        > empty lines are ignored otherwise

        Can also write something here

        The last line will be included in the show notes.""").strip()
    )


def test_generate_metadata():
    annotated_srt_file = EP0_DIR_PATH / "ep12.md"
    m4a_file = EP0_DIR_PATH / "ep12-masa.m4a"

    with TemporaryDirectory() as tmpdir:
        generate_metadata(annotated_srt_file, m4a_file, tmpdir)
        audio_out = Path(tmpdir) / "ep12-masa.mp3"
        assert audio_out.exists()
        id3 = ID3(audio_out)
        assert len(id3.getall("CHAP")) == 2
        assert "Chapter 1" in id3.pprint()
        assert "Chapter 2" in id3.pprint()
        show_notes_out = Path(tmpdir) / "ep12-masa.show_notes.md"
        show_notes_out.exists()
        chapters_out = Path(tmpdir) / "ep12-masa.chapters.txt"
        chapters_out.exists()


def test_generate_metadata_find_audio_file():
    annotated_srt_file = EP0_DIR_PATH / "ep12.md"
    input_audio_file = EP0_DIR_PATH / "ep12-masa.m4a"

    with TemporaryDirectory() as in_dir, TemporaryDirectory() as out_dir:
        in_dir, out_dir = Path(in_dir), Path(out_dir)
        shutil.copy(str(annotated_srt_file), in_dir / "ep12.md")
        convert_to_mp3(input_audio_file, in_dir / "ep12.mp3")
        generate_metadata(
            in_dir / "ep12.md", None, out_dir
        )  # No audio file specified but will find ep12.m4a
        audio_out = out_dir / "ep12.mp3"
        assert audio_out.exists()


def test_generate_metadata_no_audio_file():
    annotated_srt_file = EP0_DIR_PATH / "ep12.md"

    with TemporaryDirectory() as in_dir, TemporaryDirectory() as out_dir:
        in_dir, out_dir = Path(in_dir), Path(out_dir)
        shutil.copy(str(annotated_srt_file), in_dir / "ep12.md")
        generate_metadata(in_dir / "ep12.md", None, out_dir)  # No audio file
        chapter_out = out_dir / "ep12.chapters.txt"
        assert chapter_out.exists()


def test_generate_metadata_invalid_audio():
    annotated_srt_file = EP0_DIR_PATH / "ep12.md"

    with TemporaryDirectory() as in_dir, TemporaryDirectory() as out_dir:
        in_dir, out_dir = Path(in_dir), Path(out_dir)
        shutil.copy(str(annotated_srt_file), in_dir / "ep12.md")
        with pytest.raises(FileNotFoundError):
            generate_metadata(in_dir / "ep12.md", "/foo/var", out_dir)  # No audio file


def test_generate_metadata_invalid_source():
    with TemporaryDirectory() as in_dir, TemporaryDirectory() as out_dir:
        in_dir, out_dir = Path(in_dir), Path(out_dir)
        with pytest.raises(FileNotFoundError):
            generate_metadata(in_dir / "ep12.md", None, out_dir)  # No SRT file



import tempfile
import wave
from pathlib import Path

import pytest
from pydub import AudioSegment

from waddle.processing.combine import (
    combine_segments_into_audio,
)
from waddle.utils import format_audio_filename


def create_dummy_segments(dir_path, timeline):
    segs_folder = dir_path / "segments"
    segs_folder.mkdir(parents=True, exist_ok=True)

    for start, end in timeline:
        duration_ms = end - start
        AudioSegment.silent(duration=duration_ms).export(
            segs_folder / format_audio_filename("seg", start, end),
            format="wav",
        )
    return segs_folder


def get_wav_duration(filename):
    """Returns the duration of a WAV file."""
    with wave.open(filename, "r") as wav_file:
        frames = wav_file.getnframes()
        rate = wav_file.getframerate()
        duration = frames / float(rate)
        return duration


def test_combine_segments_into_audio_no_files():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        segs_folder = temp_dir_path / "empty_segments"
        segs_folder.mkdir(parents=True, exist_ok=True)
        output_audio_path = temp_dir_path / "output.wav"

        combine_segments_into_audio(segs_folder, output_audio_path)
        assert output_audio_path.exists(), "Output audio file was not created."

        with wave.open(str(output_audio_path), "r") as wf:
            assert wf.getnframes() > 0, "Output audio file is empty."


def test_combine_segments_into_audio():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        output_audio_path = temp_dir_path / "output.wav"
        timeline = [(0, 100), (150, 250), (250, 299)]
        segs_folder = create_dummy_segments(temp_dir_path, timeline)
        combine_segments_into_audio(segs_folder, output_audio_path)

        assert output_audio_path.exists(), "Output audio file was not created."
        assert pytest.approx(get_wav_duration(str(output_audio_path)), 0.001) == 299 / 1000


def test_combine_segments_into_audio_extra_segment():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        output_audio_path = temp_dir_path / "output.wav"
        timeline = [(100, 200), (200, 350), (501, 555)]
        segs_folder = create_dummy_segments(temp_dir_path, timeline)
        combine_segments_into_audio(segs_folder, output_audio_path)

        assert output_audio_path.exists(), "Output audio file was not created."
        assert pytest.approx(get_wav_duration(str(output_audio_path)), 0.001) == 555 / 1000


def test_combine_segments_into_audio_with_timeline():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        output_audio_path = temp_dir_path / "output.wav"
        timeline = [(0, 100), (150, 250), (250, 299)]
        segs_folder = create_dummy_segments(temp_dir_path, timeline)
        combine_segments_into_audio(segs_folder, output_audio_path, timeline)

        assert output_audio_path.exists(), "Output audio file was not created."
        assert (
            pytest.approx(get_wav_duration(str(output_audio_path)), 0.001)
            == (100 + 100 + 49) / 1000
        )


def test_combine_segments_into_audio_extra_segment_with_timeline():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        output_audio_path = temp_dir_path / "output.wav"
        timeline = [(100, 200), (200, 350), (501, 555)]
        segs_folder = create_dummy_segments(temp_dir_path, timeline)
        combine_segments_into_audio(segs_folder, output_audio_path, timeline)

        assert output_audio_path.exists(), "Output audio file was not created."
        assert (
            pytest.approx(get_wav_duration(str(output_audio_path)), 0.001)
            == (100 + 150 + 54) / 1000
        )


def test_combine_segments_into_audio_no_files_with_timeline():
    """Test handling of an empty segment folder in combine_segments_into_audio."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        segs_folder = temp_dir_path / "empty_segments"
        segs_folder.mkdir(parents=True, exist_ok=True)
        output_audio_path = temp_dir_path / "output.wav"
        timeline = [(0, 100), (150, 250)]

        combine_segments_into_audio(segs_folder, output_audio_path, timeline)

        assert output_audio_path.exists(), (
            "Output silent audio file should be created when no segments are available."
        )


def test_compare_two_combine_segments():
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_dir_path = Path(temp_dir)
        output_audio_path_0 = temp_dir_path / "output_0.wav"
        timeline = [(0, 100), (150, 250), (250, 299)]
        segs_folder = create_dummy_segments(temp_dir_path, timeline)
        combine_segments_into_audio(segs_folder, output_audio_path_0)

        assert output_audio_path_0.exists(), "Output audio file was not created."
        with wave.open(str(output_audio_path_0), "r") as wf:
            assert wf.getnframes() > 0, "Output audio file is empty."

        output_audio_path_1 = temp_dir_path / "output_1.wav"
        segs_folder = create_dummy_segments(temp_dir_path, timeline)
        new_timeline = [(timeline[0][0], timeline[-1][1])]
        combine_segments_into_audio(segs_folder, output_audio_path_1, new_timeline)

        assert output_audio_path_1.exists(), "Output audio file was not created."
        with wave.open(str(output_audio_path_1), "r") as wf:
            assert wf.getnframes() > 0, "Output audio file is empty."

        assert pytest.approx(get_wav_duration(str(output_audio_path_0)), 0.001) == get_wav_duration(
            str(output_audio_path_1)
        )
